import cbor2 as cbor
import hashlib
import json
import time
import logging
import collections
import collections.abc
import traceback
import random
import socket
import sys

# JadeError
from .jade_error import JadeError

# Low-level comms backends
from .jade_serial import JadeSerialImpl
from .jade_tcp import JadeTCPImpl

# 'jade' logger
logger = logging.getLogger(__name__)
device_logger = logging.getLogger(f'{__name__}-device')

# BLE comms backend is optional
# It relies on the BLE dependencies being available
try:
    from .jade_ble import JadeBleImpl
except (ImportError, FileNotFoundError) as e:
    logger.warning(e)
    logger.warning('BLE scanning/connectivity will not be available')


# Default serial connection
DEFAULT_BAUD_RATE = 115200
DEFAULT_SERIAL_TIMEOUT = 120

# Default BLE connection
DEFAULT_BLE_DEVICE_NAME = 'Jade'
DEFAULT_BLE_SERIAL_NUMBER = None
DEFAULT_BLE_SCAN_TIMEOUT = 60


def _hexlify(data):
    """
    Helper to map bytes-like types into hex-strings
    to make for prettier message-logging.

    Parameters
    ----------
    data : any
        The object to hexlify.
        - bytes or bytearrays have 'hex()' method invoked
        - list and dicts (values) have this function mapped over them
        - Otherwise the input is returned unchanged
    """
    if data is None:
        return None
    elif isinstance(data, bytes) or isinstance(data, bytearray):
        return data.hex()
    elif isinstance(data, list):
        return [_hexlify(item) for item in data]
    elif isinstance(data, dict):
        return {k: _hexlify(v) for k, v in data.items()}
    else:
        return data

# NOTE: Removed entirely for electrum - so it is not used silently as a fallback.
# (hard error preferred in that case)
# Jade repo api will be improved to make enabling this function more explicit
# try:
#     import requests
#
#     def _http_request(params):
#         """
#         Simple http request function which can be used when a Jade response
#         requires an external http call.
#         The default implementation used in JadeAPI._jadeRpc() below.
#         NOTE: Only available if the 'requests' dependency is available.
#
#         Callers can supply their own implementation of this call where it is required.
#
#         Parameters
#         ----------
#         data : dict
#             A dictionary structure describing the http call to make
#
#         Returns
#         -------
#         dict
#             with single key 'body', whose value is the json returned from the call
#
#         """
#         logger.debug('_http_request: {}'.format(params))
#
#         # Use the first non-onion url
#         url = [url for url in params['urls'] if not url.endswith('.onion')][0]
#
#         if params['method'] == 'GET':
#             assert 'data' not in params, 'Cannot pass body to requests.get'
#             def http_call_fn(): return requests.get(url)
#         elif params['method'] == 'POST':
#             data = json.dumps(params['data'])
#             def http_call_fn(): return requests.post(url, data)
#         else:
#             raise JadeError(1, "Only GET and POST methods supported", params['method'])
#
#         try:
#             f = http_call_fn()
#             logger.debug("http_request received reply: {}".format(f.text))
#
#             if f.status_code != 200:
#                 logger.error("http error {} : {}".format(f.status_code, f.text))
#                 raise ValueError(f.status_code)
#
#             assert params['accept'] == 'json'
#             f = f.json()
#         except Exception as e:
#             logging.error(e)
#             f = None
#
#         return {'body': f}
#
# except ImportError as e:
#     logger.info(e)
#     logger.info('Default _http_requests() function will not be available')

def generate_dump():
    while True:
        try:
            with socket.create_connection(("localhost", 4444)) as s:
                output = b""
                while b"Open On-Chip Debugger" not in output:
                    data = s.recv(1024)
                    if not data:
                        continue
                    output += data

                s.sendall(b"esp gcov dump\n")

                output = b""
                while b"Targets disconnected." not in output:
                    data = s.recv(1024)
                    if not data:
                        continue
                    output += data
                s.sendall(b"resume\n")
                time.sleep(1)
            return
        except ConnectionRefusedError:
            pass


class JadeAPI:
    """
    High-Level Jade Client API
    Builds on a JadeInterface to provide a meaningful API

    Either:
    a) use with JadeAPI.create_[serial|ble]() as jade:
    (recommended)
    or:
    b) use JadeAPI.create_[serial|ble], then call connect() before
    using, and disconnect() when finished
    (caveat cranium)
    or:
    c) use ctor to wrap existing JadeInterface instance
    (caveat cranium)
    """

    def __init__(self, jade):
        assert jade is not None
        self.jade = jade

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc, tb):
        if (exc_type):
            logger.info("Exception causing JadeAPI context exit.")
            logger.info(exc_type)
            logger.info(exc)
            traceback.print_tb(tb)
        self.disconnect(exc_type is not None)

    @staticmethod
    def create_serial(device=None, baud=None, timeout=None):
        """
        Create a JadeAPI object using the serial interface described.

        Parameters
        ----------
        device : str, optional
            The device identifier for the serial device.
            Underlying implementation will default (to /dev/ttyUSB0)

        baud : int, optional
            The communication baud rate.
            Underlying implementation will default (to 115200)

        timeout : int, optional
            The serial read timeout when awaiting messages.
            Underlying implementation will default (to 120s)

        Returns
        -------
        JadeAPI
            API object configured to use given serial parameters.
            NOTE: the api instance has not yet tried to contact the hw
            - caller must call 'connect()' before trying to use the Jade.
        """
        impl = JadeInterface.create_serial(device, baud, timeout)
        return JadeAPI(impl)

    @staticmethod
    def create_ble(device_name=None, serial_number=None,
                   scan_timeout=None, loop=None):
        """
        Create a JadeAPI object using the BLE interface described.
        NOTE: raises JadeError if BLE dependencies not installed.

        Parameters
        ----------
        device_name : str, optional
            The device name of the desired BLE device.
            Underlying implementation will default (to 'Jade')

        serial_number : int, optional
            The serial number of the desired BLE device
            - used to disambiguate multiple beacons with the same 'device name'
            Underlying implementation will connect to the first beacon it scans
            with the matching 'device name'.

        scan_timeout : int, optional
            The timeout when scanning for devices which match the device name/serial number.
            Underlying implementation will default (to 60s)

        loop : optional
            The asynchio event loop to use, if required.
            Underlying implementation will default (to asyncio.get_event_loop())

        Returns
        -------
        JadeAPI
            API object configured to use given BLE parameters.
            NOTE: the api instance has not yet tried to contact the hw
            - caller must call 'connect()' before trying to use the Jade.

        Raises
        ------
        JadeError if BLE backend not available (ie. BLE dependencies not installed)
        """
        impl = JadeInterface.create_ble(device_name, serial_number,
                                        scan_timeout, loop)
        return JadeAPI(impl)

    def connect(self):
        """
        Try to connect the underlying transport interface (eg. serial, ble, etc.)
        Raises an exception on failure.
        """
        self.jade.connect()

    def disconnect(self, drain=False):
        """
        Disconnect the underlying transport (eg. serial, ble, etc.)

        Parameters
        ----------
        drain : bool, optional
            When true log any/all remaining messages/data, otherwise silently discard.
            NOTE: can prevent disconnection if data is arriving constantly.
            Defaults to False.
        """
        self.jade.disconnect(drain)

    def drain(self):
        """
        Log any/all outstanding messages/data.
        NOTE: can run indefinitely if data is arriving constantly.
        """
        self.jade.drain()

    @staticmethod
    def _get_result_or_raise_error(reply):
        """
        Raise any error message returned from a Jade rpc call as an exception.

        Parameters
        ----------
        reply : dict
            Dictionary representing a reply from a Jade rpc call.

        Returns
        -------
        dict
            Any nested 'result' structure, if the reply is not an error.

        Raises
        ------
        JadeError
            If the reply represented an error, including all details received.
        """
        if 'error' in reply:
            e = reply['error']
            raise JadeError(e.get('code'), e.get('message'), e.get('data'))

        return reply['result']

    def _jadeRpc(self, method, params=None, inputid=None, http_request_fn=None, long_timeout=False):
        """
        Helper to make a request/reply rpc call over the underlying transport interface.
        NOTE: interface must be 'connected'.

        If the call returns an 'http_request' structure, this is handled here and the http
        call is made, and the result is passed into the rpc method given in 'on reply', by
        calling this function recursively.

        Parameters
        ----------
        method : str
            rpc method to invoke

        params : dict, optional
            any parameters to pass to the rpc method
            Defaults to None.

        inputid : str, optional
            Any specific 'id' to use in the rpc message.
            Defaults to a using a pseudo-random id generated in-situ.

        http_request_fn : function, optional
            A function which accepts a dict (containing a description of the http request), makes
            the described http call, and returns the body data in an element called 'body'.
            Defaults to _http_request() above.

        long_timeout : bool, optional
            Whether the rpc call should use an indefinitely long timeout, rather than that set on
            construction.
            (Useful if the call involves a non-trivial user interaction with the device.)
            Defaults to False.

        Returns
        -------
        dict
            The reply from the rpc call.
            NOTE: will return the last/final reply after a sequence of calls, where 'http_request'
            was returned and remote data was fetched and passed into s subsequent call.
        """
        newid = inputid if inputid else str(random.randint(100000, 999999))
        request = self.jade.build_request(newid, method, params)
        reply = self.jade.make_rpc_call(request, long_timeout)
        result = self._get_result_or_raise_error(reply)

        # The Jade can respond with a request for interaction with a remote
        # http server. This is used for interaction with the pinserver but the
        # code below acts as a dumb proxy and simply makes the http request and
        # forwards the response back to the Jade.
        # Note: the function called to make the http-request can be passed in,
        # or it can default to the simple _http_request() function above, if available.
        if isinstance(result, collections.abc.Mapping) and 'http_request' in result:
            this_module = sys.modules[__name__]
            make_http_request = http_request_fn or getattr(this_module, '_http_request', None)
            assert make_http_request, 'Default _http_request() function not available'

            http_request = result['http_request']
            http_response = make_http_request(http_request['params'])
            return self._jadeRpc(
                http_request['on-reply'],
                http_response['body'],
                http_request_fn=make_http_request,
                long_timeout=long_timeout)

        return result

    def ping(self):
        """
        RPC call to test the connection to Jade and that Jade is powered on and receiving data, and
        return whether the main task is currently handling a message, handling user menu navigation
        or is idle.

        NOTE: unlike all other calls this is not queued and handled in fifo order - this message is
        handled immediately and the response sent as quickly as possible.  This call does not block.
        If this call is made in parallel with Jade processing other messages, the replies may be
        out of order (although the message 'id' should still be correct).  Use with caution.

        Returns
        -------
        0 if the main task is currently idle
        1 if the main task is handling a client message
        2 if the main task is handling user ui menu navigation
        """
        return self._jadeRpc('ping')

    def get_version_info(self, nonblocking=False):
        """
        RPC call to fetch summary details pertaining to the hardware unit and running firmware.

        Parameters
        ----------
        nonblocking : bool
            If True message will be handled immediately (see also ping()) *experimental feature*

        Returns
        -------
        dict
            Contains keys for various info describing the hw and running fw
        """
        params = {'nonblocking': True} if nonblocking else None
        return self._jadeRpc('get_version_info', params)

    def add_entropy(self, entropy):
        """
        RPC call to add client entropy into the unit RNG entropy pool.

        Parameters
        ----------
        entropy : bytes
            Bytes to fold into the hw entropy pool.

        Returns
        -------
        bool
            True on success
        """
        params = {'entropy': entropy}
        return self._jadeRpc('add_entropy', params)

    def set_epoch(self, epoch=None):
        """
        RPC call to set the current time epoch value, required for TOTP use.
        NOTE: The time is lost on each power-down and must be reset on restart/reconnect before
        TOTP can be used.

        Parameters
        ----------
        epoch : int, optional
            Current epoch value, in seconds.  Defaults to int(time.time()) value.

        Returns
        -------
        bool
            True on success
        """
        params = {'epoch': epoch if epoch is not None else int(time.time())}
        return self._jadeRpc('set_epoch', params)

    def logout(self):
        """
        RPC call to logout of any wallet loaded on the Jade unit.
        Any key material is freed and zero'd.
        Call always returns true.

        Returns
        -------
        bool
            True
        """
        return self._jadeRpc('logout')

    def ota_update(self, fwcmp, fwlen, chunksize, fwhash=None, patchlen=None, cb=None,
                   gcov_dump=False):
        """
        RPC call to attempt to update the unit's firmware.

        Parameters
        ----------
        fwcmp : bytes
            The compressed firmware image to upload to the Jade unit.  Can be a full firmware or
            and incremental diff to be applied to the currently running firmware image.
        fwlen : int
            The size of the new complete (uncompressed) firmware image (after any delta is applied).
        chunksize : int
            The size of the chunks used to upload the compressed firmware.  Each chunk is uploaded
            and ack'd by the hw unit.
            The maximum supported chunk size is given in the version info data, under the key
            'JADE_OTA_MAX_CHUNK'.
        fwhash: 32-bytes, optional
            The sha256 hash of the full uncompressed final firmware image.  In the case of a full
            firmware upload this should be the hash of the uncompressed file.  In the case of a
            delta update this is the hash of the expected final image - ie. the existing firmware
            with the uploaded delta applied.  ie. it is a verification of the fw image Jade will try
            to boot. Optional for backward-compatibility - may become mandatory in a future release.
        patchlen: int, optional
            If the compressed firmware bytes are an incremental diff to be applied to the running
            firmware image, this is the size of that patch when uncompressed.
            Defaults to None, implying the compressed data is a full firmware image upload.
            (Compare with fwlen - the size of the final fw image.)
        cb : function, optional
            Callback function accepting two integers - the amount of compressed firmware sent thus
            far, and the total length of the compressed firmware to send.
            If passed, this function is invoked each time a fw chunk is successfully uploaded and
            ack'd by the hw, to notify of upload progress.
            Defaults to None, and nothing is called to report upload progress.

        Returns
        -------
        bool
            True if no errors were reported - on next restart the hw unit will attempt to boot the
            new firmware.
        """

        # Compute the sha256 hash of the compressed file being uploaded
        cmphasher = hashlib.sha256()
        cmphasher.update(fwcmp)
        cmphash = cmphasher.digest()
        cmplen = len(fwcmp)

        # Initiate OTA
        ota_method = 'ota'
        params = {'fwsize': fwlen,
                  'cmpsize': cmplen,
                  'cmphash': cmphash}

        if fwhash is not None:
            params['fwhash'] = fwhash

        if patchlen is not None:
            ota_method = 'ota_delta'
            params['patchsize'] = patchlen

        result = self._jadeRpc(ota_method, params)
        assert result is True

        # Write binary chunks
        written = 0
        while written < cmplen:
            remaining = cmplen - written
            length = min(remaining, chunksize)
            chunk = bytes(fwcmp[written:written + length])
            result = self._jadeRpc('ota_data', chunk)
            assert result is True
            written += length

            if (cb):
                cb(written, cmplen)

        if gcov_dump:
            self.run_remote_gcov_dump()

        # All binary data uploaded
        return self._jadeRpc('ota_complete')

    def run_remote_selfcheck(self):
        """
        RPC call to run in-built tests.
        NOTE: Only available in a DEBUG build of the firmware.

        Returns
        -------
        int
            Time in ms for the internal tests to run, as measured on the hw.
            ie. excluding any messaging overhead
        """
        return self._jadeRpc('debug_selfcheck', long_timeout=True)

    def run_remote_gcov_dump(self):
        """
        RPC call to run in-built gcov-dump.
        NOTE: Only available in a DEBUG build of the firmware.

        Returns
        -------
        bool
            Always True.
        """
        result = self._jadeRpc('debug_gcov_dump', long_timeout=True)
        time.sleep(0.5)
        generate_dump()
        time.sleep(2)
        return result

    def capture_image_data(self, check_qr=False):
        """
        RPC call to capture raw image data from the camera.
        See also scan_qr() below.
        NOTE: Only available in a DEBUG build of the firmware.

        Parameters
        ----------
        check_qr : bool, optional
            If True only images which contain a valid qr code are captured and returned.
            If False, any image is considered valid and is returned.
            Defaults to False

        Returns
        -------
        bytes
            Raw image data from the camera framebuffer
        """
        params = {'check_qr': check_qr}
        return self._jadeRpc('debug_capture_image_data', params)

    def scan_qr(self, image):
        """
        RPC call to scan a passed image and return any data extracted from any qr image.
        Exercises the camera image capture, but ignores result and uses passed image instead.
        See also capture_image_data() above.
        NOTE: Only available in a DEBUG build of the firmware.

        Parameters
        ----------
        image : bytes
            The image data (as obtained from capture_image_data() above).

        Returns
        -------
        bytes
            String or byte data obtained from the image (via qr code)
        """
        params = {'image': image}
        return self._jadeRpc('debug_scan_qr', params)

    def clean_reset(self):
        """
        RPC call to clean/reset memory and storage, as much as is practical.
        NOTE: Only available in a DEBUG build of the firmware.

        Returns
        -------
        bool
            True on success.
        """
        return self._jadeRpc('debug_clean_reset')

    def set_mnemonic(self, mnemonic, passphrase=None, temporary_wallet=False):
        """
        RPC call to set the wallet mnemonic (in RAM only - flash storage is untouched).
        NOTE: Only available in a DEBUG build of the firmware.

        Parameters
        ----------
        mnemonic : str
            The wallet mnemonic to set.

        passphrase : str, optional
            Any bip39 passphrase to apply.
            Defaults to None.

        temporary_wallet : bool, optional
            Whether to treat this wallet/mnemonic as an 'Emergency Restore' temporary wallet, as
            opposed to one successfully loaded from the flash storage.
            NOTE: in either case the wallet is only set in RAM, and flash storage is not affected.
            Defaults to False.

        Returns
        -------
        bool
            True on success.
        """
        params = {'mnemonic': mnemonic, 'passphrase': passphrase,
                  'temporary_wallet': temporary_wallet}
        return self._jadeRpc('debug_set_mnemonic', params)

    def set_seed(self, seed):
        """
        RPC call to set the wallet seed.
        NOTE: Only available in a DEBUG build of the firmware.
        NOTE: Setting a seed always sets a 'temporary' wallet.

        Parameters
        ----------
        seed : bytes
            The wallet seed to set as a temporary wallet (cannot be persisted in flash).

        Returns
        -------
        bool
            True on success.
        """
        params = {'seed': seed}
        return self._jadeRpc('debug_set_mnemonic', params)

    def get_bip85_bip39_entropy(self, num_words, index, pubkey):
        """
        RPC call to fetch encrypted bip85-bip39 entropy.
        NOTE: Only available in a DEBUG build of the firmware.

        Parameters
        ----------
        num_words : int
            The number of words the entropy is required to produce.

        index : int
            The index to use in the bip32 path to calculate the entropy.

        pubkey: 33-bytes
            The host ephemeral pubkey to use to generate a shared ecdh secret to use as an AES key
            to encrypt the returned entropy.

        Returns
        -------
        dict
            pubkey - 33-bytes, Jade's ephemeral pubkey used to generate a shared ecdh secret used as
            an AES key to encrypt the returned entropy
            encrypted - bytes, the requested bip85 bip39 entropy, AES encrypted with the first key
            derived from the ecdh shared secret, prefixed with the iv
            hmac - 32-bytes, the hmac of the encrypted buffer, using the second key derived from the
            ecdh shared secret
        """
        params = {'num_words': num_words,
                  'index': index,
                  'pubkey': pubkey}
        return self._jadeRpc('get_bip85_bip39_entropy', params)

    def set_pinserver(self, urlA=None, urlB=None, pubkey=None, cert=None):
        """
        RPC call to explicitly set (override) the details of the blind pinserver used to
        authenticate the PIN entered on the Jade unit.
        This data is recorded in the hw flash, and returned to the caller when authenticating
        (in auth_user(), below).

        Parameters
        ----------
        urlA : str, optional
            The primary url of the pinserver to use.

        urlB : str, optional
            Any secondary url of the pinserver to use.

        pubkey : bytes, optional
            The public key used to verify pinserver signed payloads.

        cert : bytes, optional
            Any additional certificate required to verify the pinserver identity.

        Returns
        -------
        bool
            True on success.
        """
        params = {}
        if urlA is not None or urlB is not None:
            params['urlA'] = urlA
            params['urlB'] = urlB
        if pubkey is not None:
            params['pubkey'] = pubkey
        if cert is not None:
            params['certificate'] = cert
        return self._jadeRpc('update_pinserver', params)

    def reset_pinserver(self, reset_details, reset_certificate):
        """
        RPC call to reset any formerly overridden pinserver details to their defaults.

        Parameters
        ----------
        reset_details : bool, optional
            If set, any overridden urls and pubkey are reset to their defaults.

        reset_certificate : bool, optional
            If set, any additional certificate is reset (to None).

        Returns
        -------
        bool
            True on success.
        """
        params = {'reset_details': reset_details,
                  'reset_certificate': reset_certificate}
        return self._jadeRpc('update_pinserver', params)

    def auth_user(self, network, http_request_fn=None, epoch=None):
        """
        RPC call to authenticate the user on the hw device, for using with the network provided.

        Parameters
        ----------
        network : str
            The name of the network intended for use - eg. 'mainnet', 'liquid', 'testnet' etc.
            This is verified against the networks allowed on the hardware.

        http_request_fn : function, optional
            Optional http-request function to pass http requests to the Jade pinserver.
            Default behaviour is to use the '_http_request()' function which defers to the
            'requests' module.
            If the 'reqests' module is not available, no default http-request function is created,
            and one must be supplied here.

        epoch : int, optional
            Current epoch value, in seconds.  Defaults to int(time.time()) value.

        Returns
        -------
        bool
            True is returned immediately if the hw is already unlocked for use on the given network.
            True if the PIN is entered and verified with the remote blind pinserver.
            False if the PIN entered was incorrect.
        """
        params = {'network': network, 'epoch': epoch if epoch is not None else int(time.time())}
        return self._jadeRpc('auth_user', params,
                             http_request_fn=http_request_fn,
                             long_timeout=True)

    def register_otp(self, otp_name, otp_uri):
        """
        RPC call to register a new OTP record on the hw device.

        Parameters
        ----------
        otp_name : str
            An identifying name for this OTP record

        otp_uri : str
            The uri of this OTP record - must begin 'otpauth://'

        Returns
        -------
        bool
            True if the OTP uri was validated and persisted on the hw
        """
        params = {'name': otp_name, 'uri': otp_uri}
        return self._jadeRpc('register_otp', params)

    def get_otp_code(self, otp_name, value_override=None):
        """
        RPC call to fetch a new OTP code from the hw device.

        Parameters
        ----------
        otp_name : str
            An identifying name for the OTP record to use

        value_override : int
            An overriding HOTP counter or TOTP timestamp to use.
            NOTE: Only available in a DEBUG build of the firmware.

        Returns
        -------
        bool
            True if the OTP uri was validated and persisted on the hw
        """
        params = {'name': otp_name}
        if value_override is not None:
            params['override'] = value_override
        return self._jadeRpc('get_otp_code', params)

    def get_xpub(self, network, path):
        """
        RPC call to fetch an xpub for the given bip32 path for the given network.

        Parameters
        ----------
        network : str
            Network to which the xpub applies - eg. 'mainnet', 'liquid', 'testnet', etc.

        path : [int]
            bip32 path for which the xpub should be generated.

        Returns
        -------
        str
            base58 encoded xpub
        """
        params = {'network': network, 'path': path}
        return self._jadeRpc('get_xpub', params)

    def get_registered_multisigs(self):
        """
        RPC call to fetch brief summaries of any multisig wallets registered to this signer.

        Returns
        -------
        dict
            Brief description of registered multisigs, keyed by registration name.
            Each entry contains keys:
                variant - str, script type, eg. 'sh(wsh(multi(k)))'
                sorted - boolean, whether bip67 key sorting is applied
                threshold - int, number of signers required,N
                num_signers - total number of signatories, M
                master_blinding_key - 32-bytes, any liquid master blinding key for this wallet
        """
        return self._jadeRpc('get_registered_multisigs')

    def get_registered_multisig(self, multisig_name, as_file=False):
        """
        RPC call to fetch details of a named multisig wallet registered to this signer.
        NOTE: the multisig wallet must have been registered with firmware v1.0.23 or later
        for the full signer details to be persisted and available.

        Parameters
        ----------
        multisig_name : string
            Name of multsig registration record to return.

        as_file : string, optional
            If true the flat file format is returned, otherwise structured json is returned.
            Defaults to false.

        Returns
        -------
        dict
            Description of registered multisig wallet identified by registration name.
            Contains keys:
                is_file is true:
                    multisig_file - str, the multisig file as produced by several wallet apps.
                    eg:
                        Name: MainWallet
                        Policy: 2 of 3
                        Format: P2WSH
                        Derivation: m/48'/0'/0'/2'

                        B237FE9D: xpub6E8C7BX4c7qfTsX7urnXggcAyFuhDmYLQhwRwZGLD9maUGWPinuc9k96ej...
                        249192D2: xpub6EbXynW6xjYR3crcztum6KzSWqDJoAJQoovwamwVnLaCSHA6syXKPnJo6U...
                        67F90FFC: xpub6EHuWWrYd8bp5FS1XAZsMPkmCqLSjpULmygWqAqWRCCjSWQwz6ntq5KnuQ...

                is_file is false:
                    multisig_name - str, name of multisig registration
                    variant - str, script type, eg. 'sh(wsh(multi(k)))'
                    sorted - boolean, whether bip67 key sorting is applied
                    threshold - int, number of signers required,N
                    master_blinding_key - 32-bytes, any liquid master blinding key for this wallet
                    signers - dict containing keys:
                        fingerprint - 4 bytes, origin fingerprint
                        derivation - [int], bip32 path from origin to signer xpub provided
                        xpub - str, base58 xpub of signer
                        path - [int], any fixed path to always apply after the xpub - usually empty.

        """
        params = {'multisig_name': multisig_name,
                  'as_file': as_file}
        return self._jadeRpc('get_registered_multisig', params)

    def register_multisig(self, network, multisig_name, variant, sorted_keys, threshold, signers,
                          master_blinding_key=None):
        """
        RPC call to register a new multisig wallet, which must contain the hw signer.
        A registration name is provided - if it already exists that record is overwritten.

        Parameters
        ----------
        network : string
            Network to which the multisig should apply - eg. 'mainnet', 'liquid', 'testnet', etc.

        multisig_name : string
            Name to use to identify this multisig wallet registration record.
            If a registration record exists with the name given, that record is overwritten.

        variant : str
            The script type - one of 'sh(multi(k))', 'wsh(multi(k))', 'sh(wsh(multi(k)))'

        sorted_keys : bool
            Whether this is a 'sortedmulti()' wallet - ie. whether to apply bip67 sorting to the
            pubkeys when generating redeem scripts.

        threshold : int
            Number of signers required.

        signers : [dict]
            Description of signers - should include keys:
                - 'fingerprint' - 4 bytes, origin fingerprint
                - 'derivation' - [int], bip32 path from origin to signer xpub provided
                - 'xpub' - str, base58 xpub of signer - will be verified for hw unit signer
                - 'path' - [int], any fixed path to always apply after the xpub - usually empty.

        master_blinding_key : 32-bytes, optional
            The master blinding key to use for this multisig wallet on liquid.
            Optional, defaults to None.
            Logically mandatory when 'network' indicates a liquid network and the Jade is to be
            used to generate confidential addresses, blinding keys, blinding nonces, asset blinding
            factors or output commitments.

        Returns
        -------
        bool
            True on success, implying the mutisig wallet can now be used.
        """
        params = {'network': network, 'multisig_name': multisig_name,
                  'descriptor': {'variant': variant, 'sorted': sorted_keys,
                                 'threshold': threshold, 'signers': signers,
                                 'master_blinding_key': master_blinding_key}}
        return self._jadeRpc('register_multisig', params)

    def register_multisig_file(self, multisig_file):
        """
        RPC call to register a new multisig wallet, which must contain the hw signer.
        A registration file is provided - as produced my several wallet apps.

        Parameters
        ----------
        multisig_file : string
            The multisig file as produced by several wallet apps.
            eg:
                Name: MainWallet
                Policy: 2 of 3
                Format: P2WSH
                Derivation: m/48'/0'/0'/2'

                B237FE9D: xpub6E8C7BX4c7qfTsX7urnXggcAyFuhDmYLQhwRwZGLD9maUGWPinuc9k96ejhEQ1DCk...
                249192D2: xpub6EbXynW6xjYR3crcztum6KzSWqDJoAJQoovwamwVnLaCSHA6syXKPnJo6U3bVeGde...
                67F90FFC: xpub6EHuWWrYd8bp5FS1XAZsMPkmCqLSjpULmygWqAqWRCCjSWQwz6ntq5KnuQnL23No2...

    Returns
    -------
    bool
        True on success, implying the mutisig wallet can now be used.
    """
        params = {'multisig_file': multisig_file}
        return self._jadeRpc('register_multisig', params)

    def get_registered_descriptors(self):
        """
        RPC call to fetch brief summaries of any descriptor wallets registered to this signer.

        Returns
        -------
        dict
            Brief description of registered descriptor, keyed by registration name.
            Each entry contains keys:
                descriptor_len - int, length of descriptor output script
                num_datavalues - int, total number of substitution placeholders passed with script
                master_blinding_key - 32-bytes, any liquid master blinding key for this wallet
        """
        return self._jadeRpc('get_registered_descriptors')

    def get_registered_descriptor(self, descriptor_name):
        """
        RPC call to fetch details of a named descriptor wallet registered to this signer.

        Parameters
        ----------
        descriptor_name : string
            Name of descriptor registration record to return.

        Returns
        -------
        dict
            Description of registered descriptor wallet identified by registration name.
            Contains keys:
                descriptor_name - str, name of descriptor registration
                descriptor - str, descriptor output script, may contain substitution placeholders
                datavalues - dict containing placeholders for substitution into script
        """
        params = {'descriptor_name': descriptor_name}
        return self._jadeRpc('get_registered_descriptor', params)

    def register_descriptor(self, network, descriptor_name, descriptor_script, datavalues=None):
        """
        RPC call to register a new descriptor wallet, which must contain the hw signer.
        A registration name is provided - if it already exists that record is overwritten.

        Parameters
        ----------
        network : string
            Network to which the descriptor should apply - eg. 'mainnet', 'liquid', 'testnet', etc.

        descriptor_name : string
            Name to use to identify this descriptor wallet registration record.
            If a registration record exists with the name given, that record is overwritten.

        Returns
        -------
        bool
            True on success, implying the descriptor wallet can now be used.
        """
        params = {'network': network, 'descriptor_name': descriptor_name,
                  'descriptor': descriptor_script, 'datavalues': datavalues}
        return self._jadeRpc('register_descriptor', params)

    def get_receive_address(self, *args, recovery_xpub=None, csv_blocks=0,
                            variant=None, multisig_name=None, descriptor_name=None,
                            confidential=None):
        """
        RPC call to generate, show, and return an address for the given path.
        The call has three forms.

        Parameters
        ----------
        network: str
            Network to which the address should apply - eg. 'mainnet', 'liquid', 'testnet', etc.

        Then either:

        1. Blockstream Green (multisig shield) addresses
            subaccount : int
                Blockstream Green subaccount

            branch : int
                Blockstream Green derivation branch

            pointer : int
                Blockstream Green address pointer

            recovery_xpub : str, optional
                xpub of recovery key for 2of3 subaccounts.  Otherwise should be omitted.
                Defaults to None (ie. not a 2of3 subaccount).

            csv_blocks : int, optional
                Number of blocks to include in csv redeem script, if this is a csv-enabled account.
                Otherwise should be omitted.
                Defaults to 0 (ie. does not apply/not a csv-enabled account.)

        2. Generic single-sig addresses
            path: [int]
                bip32 path for which the xpub should be generated.

            variant: str
                The script type - one of 'pkh(k)', 'wpkh(k)', 'sh(wpkh(k))'

        3. Generic multisig addresses
            paths: [[int]]
                bip32 path suffixes, one for each signer, applied as a suffix to the registered
                signer path. Usually these path suffixes will all be identical.

            multisig_name : str
                The name of the registered multisig wallet record used to generate the address.

        4. Descriptor wallet addresses
            branch : int
                Multi-path derivation branch, usually 0.

            pointer : int
                Path index to descriptor

            descriptor_name : str
                The name of the registered descriptor wallet record used to generate the address.

        Returns
        -------
        str
            The address generated for the given parameters.

        """
        if multisig_name is not None:
            assert len(args) == 2
            keys = ['network', 'paths', 'multisig_name']
            args += (multisig_name,)
        elif descriptor_name is not None:
            assert len(args) == 3
            keys = ['network', 'branch', 'pointer', 'descriptor_name']
            args += (descriptor_name,)
        elif variant is not None:
            assert len(args) == 2
            keys = ['network', 'path', 'variant']
            args += (variant,)
        else:
            assert len(args) == 4
            keys = ['network', 'subaccount', 'branch', 'pointer', 'recovery_xpub', 'csv_blocks']
            args += (recovery_xpub, csv_blocks)

        params = dict(zip(keys, args))
        if confidential is not None:
            params['confidential'] = confidential

        return self._jadeRpc('get_receive_address', params)

    def sign_message(self, path, message, use_ae_signatures=False,
                     ae_host_commitment=None, ae_host_entropy=None):
        """
        RPC call to format and sign the given message, using the given bip32 path.
        Supports RFC6979 and anti-exfil signatures.

        Parameters
        ----------
        path : [int]
            bip32 path for which the signature should be generated.

        message : str
            Message string to format and sign.

        ae_host_commitment : 32-bytes, optional
            The host-commitment to use for Antil-Exfil signatures

        ae_host_entropy : 32-bytes, optional
            The host-entropy to use for Antil-Exfil signatures

        Returns
        -------
        1. Legacy/RFC6979 signatures
        str
            base64-encoded signature

        2. Anti-exfil signatures
        (bytes, str)
            signer-commitment, base64-encoded signature
        """
        if use_ae_signatures:
            # Anti-exfil protocol:
            # We send the signing request and receive the signer-commitment in
            # reply once the user confirms.
            # We can then request the actual signature passing the ae-entropy.
            params = {'path': path, 'message': message, 'ae_host_commitment': ae_host_commitment}
            signer_commitment = self._jadeRpc('sign_message', params)
            params = {'ae_host_entropy': ae_host_entropy}
            signature = self._jadeRpc('get_signature', params)
            return signer_commitment, signature
        else:
            # Standard EC signature, simple case
            params = {'path': path, 'message': message}
            return self._jadeRpc('sign_message', params)

    def sign_message_file(self, message_file):
        """
        RPC call to format and sign the given message, using the given bip32 path.
        A message file is provided - as produced by eg. Specter wallet.
        Supports RFC6979 only.

        Parameters
        ----------
        message_file : str
            Message file to parse and produce signature for.
            eg:  'signmessage m/84h/0h/0h/0/0 ascii:this is a test message'

        Returns
        -------
        str
            base64-encoded RFC6979 signature
        """
        params = {'message_file': message_file}
        return self._jadeRpc('sign_message', params)

    def get_identity_pubkey(self, identity, curve, key_type, index=0):
        """
        RPC call to fetch a pubkey for the given identity (slip13/slip17).
        NOTE: this api returns an uncompressed public key

        Parameters
        ----------
        identity : str
            Identity string to format and sign. For example ssh://satoshi@bitcoin.org

        curve : str
            Name of curve to use - currently only 'nist256p1' is supported

        key_type : str
            Key derivation type - must be either 'slip-0013' for an identity pubkey, or 'slip-0017'
            for an ecdh pubkey.

        index : int, optional
            Index number (if require multiple keys/sigs per identity)
            Defaults to 0

        Returns
        -------
        65-bytes
            Uncompressed public key for the given identity and index.
            Consistent with 'sign_identity' or 'get_identity_shared_key', depending on the
            'key_type'.

        """
        params = {'identity': identity, 'curve': curve, 'type': key_type, 'index': index}
        return self._jadeRpc('get_identity_pubkey', params)

    def get_identity_shared_key(self, identity, curve, their_pubkey, index=0):
        """
        RPC call to fetch a SLIP-0017 shared ecdh key for the identity and counterparty public key.
        NOTE: this api takes an uncompressed public key

        Parameters
        ----------
        identity : str
            Identity string to format and sign. For example ssh://satoshi@bitcoin.org

        curve : str
            Name of curve to use - currently only 'nist256p1' is supported

        their_pubkey : 65-bytes
            The counterparty's uncompressed public key

        index : int, optional
            Index number (if require multiple keys/sigs per identity)
            Defaults to 0

        Returns
        -------
        32-bytes
            The shared ecdh key for the given identity and cpty public key
            Consistent with 'get_identity_pubkey' with 'key_type=slip-0017'
        """
        params = {'identity': identity, 'curve': curve, 'index': index,
                  'their_pubkey': their_pubkey}
        return self._jadeRpc('get_identity_shared_key', params)

    def sign_identity(self, identity, curve, challenge, index=0):
        """
        RPC call to authenticate the given identity through a challenge.
        Supports RFC6979.
        Returns the signature and the associated SLIP-0013 pubkey
        NOTE: this api returns an uncompressed public key

        Parameters
        ----------
        identity : str
            Identity string to format and sign. For example ssh://satoshi@bitcoin.org

        curve : str
            Name of curve to use - currently only 'nist256p1' is supported

        challenge : bytes
            Challenge bytes to sign

        index : int, optional
            Index number (if require multiple keys/sigs per identity)
            Defaults to 0

        Returns
        -------
        dict
            Contains keys:
            pubkey - 65-bytes, the uncompressed SLIP-0013 public key, consistent with
            'get_identity_pubkey' with 'key_type=slip-0013'
            signature - 65-bytes, RFC6979 deterministic signature, prefixed with 0x00
        """
        params = {'identity': identity, 'curve': curve, 'index': index, 'challenge': challenge}
        return self._jadeRpc('sign_identity', params)

    def sign_attestation(self, challenge):
        """
        RPC call to sign passed challenge with embedded hw RSA-4096 key, such that the caller
        can check the authenticity of the hardware unit.  eg. whether it is a genuine
        Blockstream production Jade unit.
        Caller must have the public key of the external verifying authority they wish to validate
        against (eg. Blockstream's Jade verification public key).
        NOTE: only supported by ESP32S3-based hardware units.

        Parameters
        ----------
        challenge : bytes
            Challenge bytes to sign

        Returns
        -------
        dict
            Contains keys:
            signature - 512-bytes, hardware RSA signature of the SHA256 hash of the passed
                        challenge bytes.
            pubkey_pem - str, PEM export of RSA pubkey of the hardware unit, to verify the returned
            RSA signature.
            ext_signature - bytes, RSA signature of the verifying authority over the returned
            pubkey_pem data.
            (Caller can verify this signature with the public key of the verifying authority.)
        """
        params = {'challenge': challenge}
        return self._jadeRpc('sign_attestation', params)

    def get_master_blinding_key(self, only_if_silent=False):
        """
        RPC call to fetch the master (SLIP-077) blinding key for the hw signer.
        May block temporarily to request the user's permission to export.  Passing 'only_if_silent'
        causes the call to return the 'denied' error if it would normally ask the user.
        NOTE: the master blinding key of any registered multisig wallets can be obtained from
        the result of `get_registered_multisigs()`.

        Parameters
        ----------
        only_if_silent : boolean, optional
            If True Jade will return the denied error if it would normally ask the user's permission
            to export the master blinding key.  Passing False (or letting default) may block while
            asking the user to confirm the export on Jade.

        Returns
        -------
        32-bytes
            SLIP-077 master blinding key
        """
        params = {'only_if_silent': only_if_silent}
        return self._jadeRpc('get_master_blinding_key', params)

    def get_blinding_key(self, script, multisig_name=None):
        """
        RPC call to fetch the public blinding key for the hw signer.

        Parameters
        ----------
        script : bytes
            The script for which the public blinding key is required.

        multisig_name : str, optional
            The name of any registered multisig wallet for which to fetch the blinding key.
            Defaults to None

        Returns
        -------
        33-bytes
            Public blinding key for the passed script.
        """
        params = {'script': script, 'multisig_name': multisig_name}
        return self._jadeRpc('get_blinding_key', params)

    def get_shared_nonce(self, script, their_pubkey, include_pubkey=False, multisig_name=None):
        """
        RPC call to get the shared secret to unblind a tx, given the receiving script and
        the pubkey of the sender (sometimes called "blinding nonce" in Liquid).
        Optionally fetch the hw signer's public blinding key also.

        Parameters
        ----------
        script : bytes
            The script for which the blinding nonce is required.

        their_pubkey : 33-bytes
            The counterparty public key.

        include_pubkey : bool, optional
            Whether to also return the wallet's public blinding key.
            Defaults to False.

        multisig_name : str, optional
            The name of any registered multisig wallet for which to fetch the blinding nonce.
            Defaults to None

        Returns
        -------
        1. include_pubkey is False
        33-bytes
            Public blinding nonce for the passed script and counterparty public key.

        2. include_pubkey is True
        dict
            Contains keys:
            shared_nonce - 32-bytes, public blinding nonce for the passed script as above.
            blinding_key - 33-bytes, public blinding key for the passed script.
        """
        params = {'script': script, 'their_pubkey': their_pubkey,
                  'include_pubkey': include_pubkey, 'multisig_name': multisig_name}
        return self._jadeRpc('get_shared_nonce', params)

    def get_blinding_factor(self, hash_prevouts, output_index, bftype, multisig_name=None):
        """
        RPC call to get deterministic blinding factors to blind an output.
        Predicated on the host calculating the 'hash_prevouts' value correctly.
        Can fetch abf, vbf, or both together.

        Parameters
        ----------

        hash_prevouts : 32-bytes
            This value should be computed by the host as specified in bip143.
            It is not verified by Jade, since at this point Jade does not have the tx in question.

        output_index : int
            The index of the output we are trying to blind

        bftype : str
            Can be "ASSET", "VALUE", or "ASSET_AND_VALUE", to generate abf, vbf, or both.

        multisig_name : str, optional
            The name of any registered multisig wallet for which to fetch the blinding factor.
            Defaults to None

        Returns
        -------
        32-bytes or 64-bytes
            The blinding factor for "ASSET" and "VALUE" requests, or both concatenated abf|vbf
            ie. the first 32 bytes being abf, the second 32 bytes being vbf.
        """
        params = {'hash_prevouts': hash_prevouts,
                  'output_index': output_index,
                  'type': bftype,
                  'multisig_name': multisig_name}
        return self._jadeRpc('get_blinding_factor', params)

    def get_commitments(self,
                        asset_id,
                        value,
                        hash_prevouts,
                        output_index,
                        vbf=None,
                        multisig_name=None):
        """
        RPC call to generate deterministic blinding factors and commitments for a given output.
        Can optionally get a "custom" VBF, normally used for the last input where the vbf is not
        computed here, but generated on the host according to all the other values.
        The commitments generated here should be passed back into `sign_liquid_tx()`.

        Parameters
        ----------
        asset_id : 32-bytes
            asset_id as usually displayed - ie. reversed compared to network/consensus order

        value : int
            value in 'satoshi' or equivalent atomic integral unit

        hash_prevouts : 32-bytes
            This value is computed as specified in bip143.
            It is verified immediately since at this point Jade doesn't have the tx in question.
            It will be checked later during `sign_liquid_tx()`.

        output_index : int
            The index of the output we are trying to blind

        vbf : 32-bytes, optional
            The vbf to use, in preference to deterministically generating one in this call.

        multisig_name : str, optional
            The name of any registered multisig wallet for which to fetch the blinding factor.
            Defaults to None

        Returns
        -------
        dict
            Containing the blinding factors and output commitments.
        """
        params = {'asset_id': asset_id,
                  'value': value,
                  'hash_prevouts': hash_prevouts,
                  'output_index': output_index,
                  'vbf': vbf,
                  'multisig_name': multisig_name}
        return self._jadeRpc('get_commitments', params)

    def _send_tx_inputs(self, base_id, inputs, use_ae_signatures):
        """
        Helper call to send the tx inputs to Jade for signing.
        Handles legacy RFC6979 signatures, as well as the Anti-Exfil protocol.

        Parameters
        ----------
        base_id : int
            The ids of the messages sent will be increments from this base id.

        inputs : [dict]
            The tx inputs - see `sign_tx()` / `sign_liquid_tx()` for details.

        use_ae_signatures : bool
            Whether to use the anti-exfil protocol to generate the signatures

        Returns
        -------
        1. if use_ae_signatures is False
        [bytes]
            An array of signatures corresponding to the array of inputs passed.
            The signatures are in DER format with the sighash appended.
            'None' placeholder elements are used for inputs not requiring a signature.

        2. if use_ae_signatures is True
        [(32-bytes, bytes)]
            An array of pairs of signer-commitments and signatures corresponding to the inputs.
            The signatures are in DER format with the sighash appended.
            (None, None) placeholder elements are used for inputs not requiring a signature.
        """
        if use_ae_signatures:
            # Anti-exfil protocol:
            # We send one message per input (which includes host-commitment *but
            # not* the host entropy) and receive the signer-commitment in reply.
            # Once all n input messages are sent, we can request the actual signatures
            # (as the user has a chance to confirm/cancel at this point).
            # We request the signatures passing the ae-entropy for each one.
            # Send inputs one at a time, receiving 'signer-commitment' in reply
            signer_commitments = []
            host_ae_entropy_values = []
            for txinput in inputs:
                # ae-protocol - do not send the host entropy immediately
                txinput = txinput.copy() if txinput else {}  # shallow copy
                host_ae_entropy_values.append(txinput.pop('ae_host_entropy', None))

                base_id += 1
                input_id = str(base_id)
                reply = self._jadeRpc('tx_input', txinput, input_id)
                signer_commitments.append(reply)

            # Request the signatures one at a time, sending the entropy
            signatures = []
            for (i, host_ae_entropy) in enumerate(host_ae_entropy_values, 1):
                base_id += 1
                sig_id = str(base_id)
                params = {'ae_host_entropy': host_ae_entropy}
                reply = self._jadeRpc('get_signature', params, sig_id)
                signatures.append(reply)

            assert len(signatures) == len(inputs)
            return list(zip(signer_commitments, signatures))
        else:
            # Legacy protocol:
            # We send one message per input - without expecting replies.
            # Once all n input messages are sent, the hw then sends all n replies
            # (as the user has a chance to confirm/cancel at this point).
            # Then receive all n replies for the n signatures.
            # NOTE: *NOT* a sequence of n blocking rpc calls.
            # NOTE: at some point this flow should be removed in favour of the one
            # above, albeit without passing anti-exfil entropy or commitment data.

            # Send all n inputs
            requests = []
            for txinput in inputs:
                if txinput is None:
                    txinput = {}

                base_id += 1
                msg_id = str(base_id)
                request = self.jade.build_request(msg_id, 'tx_input', txinput)
                self.jade.write_request(request)
                requests.append(request)
                time.sleep(0.1)

            # Receive all n signatures
            signatures = []
            for request in requests:
                reply = self.jade.read_response()
                self.jade.validate_reply(request, reply)
                signature = self._get_result_or_raise_error(reply)
                signatures.append(signature)

            assert len(signatures) == len(inputs)
            return signatures

    def sign_liquid_tx(self, network, txn, inputs, commitments, change, use_ae_signatures=False,
                       asset_info=None, additional_info=None):
        """
        RPC call to sign a liquid transaction.

        Parameters
        ----------
        network : str
            Network to which the txn should apply - eg. 'liquid', 'liquid-testnet', etc.

        txn : bytes
            The transaction to sign

        inputs : [dict]
            The tx inputs.
                If signing this input, should contain keys:
                is_witness, bool - whether this is a segwit input
                script, bytes- the redeem script
                path, [int] - the bip32 path to sign with
                value_commitment, 33-bytes - The value commitment of the input

                This is optional if signing this input:
                sighash, int - The sighash to use, defaults to 0x01 (SIGHASH_ALL)

                These are only required for Anti-Exfil signatures:
                ae_host_commitment, 32-bytes - The host-commitment for Anti-Exfil signatures
                ae_host_entropy, 32-bytes - The host-entropy for Anti-Exfil signatures

                These are only required for advanced transactions, eg. swaps, and only when the
                inputs need unblinding.
                Not needed for vanilla send-payment/redeposit etc:
                abf, 32-bytes - asset blinding factor
                asset_id, 32-bytes - the unblinded asset-id
                asset_generator, 33-bytes - the (blinded) asset-generator
                vbf, 32-bytes - the value blinding factor
                value, int - the unblinded sats value of the input

                If not signing this input a null or an empty dict can be passed.

        commitments : [dict]
            An array sized for the number of outputs.
            Unblinded outputs should have a 'null' placeholder element.
            The commitments as retrieved from `get_commitments()`, with the addition of:
                'blinding_key', <bytes> - the output's public blinding key
                    (as retrieved from `get_blinding_key()`)

        change : [dict]
            An array sized for the number of outputs.
            Outputs which are not to this wallet should have a 'null' placeholder element.
            The output scripts for the elements with data will be verified by Jade.
            Unless the element also contains 'is_change': False, these outputs will automatically
            be approved and not be verified by the user.
            Populated elements should contain sufficient data to generate the wallet address.
            See `get_receive_address()`

        use_ae_signatures : bool, optional
            Whether to use the anti-exfil protocol to generate the signatures.
            Defaults to False.

        asset_info : [dict], optional
            Any asset-registry data relevant to the assets being transacted, such that Jade can
            display a meaningful name, issuer, ticker etc. rather than just asset-id.
            At the very least must contain 'asset_id', 'contract' and 'issuance_prevout' items,
            exactly as in the registry data.  NOTE: asset_info for the network policy-asset is
            not required.
            Defaults to None.

        additional_info: dict, optional
            Extra data about the transaction.  Only required for advanced transactions, eg. swaps.
            Not needed for vanilla send-payment/redeposit etc:
            tx_type, str: 'swap' indicates the tx represents an asset-swap proposal or transaction.
            wallet_input_summary, dict:  a list of entries containing 'asset_id' (32-bytes) and
            'satoshi' (int) showing net movement of assets out of the wallet (ie. sum of wallet
            inputs per asset, minus any change outputs).
            wallet_output_summary, dict:  a list of entries containing 'asset_id' (32-bytes) and
            'satoshi' (int) showing net movement of assets into the wallet (ie. sum of wallet
            outputs per asset, excluding any change outputs).

        Returns
        -------
        1. if use_ae_signatures is False
        [bytes]
            An array of signatures corresponding to the array of inputs passed.
            The signatures are in DER format with the sighash appended.
            'None' placeholder elements are used for inputs not requiring a signature.

        2. if use_ae_signatures is True
        [(32-bytes, bytes)]
            An array of pairs of signer-commitments and signatures corresponding to the inputs.
            The signatures are in DER format with the sighash appended.
            (None, None) placeholder elements are used for inputs not requiring a signature.
        """
        # 1st message contains txn and number of inputs we are going to send.
        # Reply ok if that corresponds to the expected number of inputs (n).
        base_id = 100 * random.randint(1000, 9999)
        params = {'network': network,
                  'txn': txn,
                  'num_inputs': len(inputs),
                  'trusted_commitments': commitments,
                  'use_ae_signatures': use_ae_signatures,
                  'change': change,
                  'asset_info': asset_info,
                  'additional_info': additional_info}

        reply = self._jadeRpc('sign_liquid_tx', params, str(base_id))
        assert reply

        # Send inputs and receive signatures
        return self._send_tx_inputs(base_id, inputs, use_ae_signatures)

    def sign_tx(self, network, txn, inputs, change, use_ae_signatures=False):
        """
        RPC call to sign a btc transaction.

        Parameters
        ----------
        network : str
            Network to which the txn should apply - eg. 'mainnet', 'testnet', etc.

        txn : bytes
            The transaction to sign

        inputs : [dict]
            The tx inputs.   Should contain keys:
                One of these is required:
                input_tx, bytes - The prior transaction which created the utxo of this input
                satoshi, int - The satoshi amount of this input - can be used in place of
                    'input_tx' for a tx with a single segwit input

                These are only required if signing this input:
                is_witness, bool - whether this is a segwit input
                script, bytes- the redeem script
                path, [int] - the bip32 path to sign with

                This is optional if signing this input:
                sighash, int - The sighash to use, defaults to 0x01 (SIGHASH_ALL)

                These are only required for Anti-Exfil signatures:
                ae_host_commitment, 32-bytes - The host-commitment for Anti-Exfil signatures
                ae_host_entropy, 32-bytes - The host-entropy for Anti-Exfil signatures

        change : [dict]
            An array sized for the number of outputs.
            Outputs which are not to this wallet should have a 'null' placeholder element.
            The output scripts for the elements with data will be verified by Jade.
            Unless the element also contains 'is_change': False, these outputs will automatically
            be approved and not be verified by the user.
            Populated elements should contain sufficient data to generate the wallet address.
            See `get_receive_address()`

        use_ae_signatures : bool
            Whether to use the anti-exfil protocol to generate the signatures

        Returns
        -------
        1. if use_ae_signatures is False
        [bytes]
            An array of signatures corresponding to the array of inputs passed.
            The signatures are in DER format with the sighash appended.
            'None' placeholder elements are used for inputs not requiring a signature.

        2. if use_ae_signatures is True
        [(32-bytes, bytes)]
            An array of pairs of signer-commitments and signatures corresponding to the inputs.
            The signatures are in DER format with the sighash appended.
            (None, None) placeholder elements are used for inputs not requiring a signature.
        """
        # 1st message contains txn and number of inputs we are going to send.
        # Reply ok if that corresponds to the expected number of inputs (n).
        base_id = 100 * random.randint(1000, 9999)
        params = {'network': network,
                  'txn': txn,
                  'num_inputs': len(inputs),
                  'use_ae_signatures': use_ae_signatures,
                  'change': change}

        reply = self._jadeRpc('sign_tx', params, str(base_id))
        assert reply

        # Send inputs and receive signatures
        return self._send_tx_inputs(base_id, inputs, use_ae_signatures)

    def sign_psbt(self, network, psbt):
        """
        RPC call to sign a passed psbt as required

        Parameters
        ----------
        network : str
            Network to which the txn should apply - eg. 'mainnet', 'testnet', etc.

        psbt : bytes
            The psbt formatted as bytes

        Returns
        -------
        bytes
            The psbt, updated with any signatures required from the hw signer
        """
        # Send PSBT message
        params = {'network': network, 'psbt': psbt}
        msgid = str(random.randint(100000, 999999))
        request = self.jade.build_request(msgid, 'sign_psbt', params)
        self.jade.write_request(request)

        # Read replies until we have them all, collate data and return.
        # NOTE: we send 'get_extended_data' messages to request more 'chunks' of the reply data.
        psbt_out = bytearray()
        while True:
            reply = self.jade.read_response()
            self.jade.validate_reply(request, reply)
            psbt_out.extend(self._get_result_or_raise_error(reply))

            if 'seqnum' not in reply or reply['seqnum'] == reply['seqlen']:
                break

            newid = str(random.randint(100000, 999999))
            params = {'origid': msgid,
                      'orig': 'sign_psbt',
                      'seqnum': reply['seqnum'] + 1,
                      'seqlen': reply['seqlen']}
            request = self.jade.build_request(newid, 'get_extended_data', params)
            self.jade.write_request(request)

        return psbt_out


class JadeInterface:
    """
    Mid-level interface to Jade
    Wraps either a serial or a ble connection
    Calls to send and receive bytes and cbor messages over the interface.

    Either:
     a) use wrapped with JadeAPI
    (recommended)
    or:
     b) use with JadeInterface.create_[serial|ble]() as jade:
          ...
    or:
     c) use JadeInterface.create_[serial|ble], then call connect() before
        using, and disconnect() when finished
    (caveat cranium)
    or:
     d) use ctor to wrap existing low-level implementation instance
    (caveat cranium)
    """

    def __init__(self, impl):
        assert impl is not None
        self.impl = impl

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc, tb):
        if (exc_type):
            logger.info("Exception causing JadeInterface context exit.")
            logger.info(exc_type)
            logger.info(exc)
            traceback.print_tb(tb)
        self.disconnect(exc_type is not None)

    @staticmethod
    def create_serial(device=None, baud=None, timeout=None):
        """
        Create a JadeInterface object using the serial interface described.

        Parameters
        ----------
        device : str, optional
            The device identifier for the serial device.
            Underlying implementation will default (to /dev/ttyUSB0)

        baud : int, optional
            The communication baud rate.
            Underlying implementation will default (to 115200)

        timeout : int, optional
            The serial read timeout when awaiting messages.
            Underlying implementation will default (to 120s)

        Returns
        -------
        JadeInterface
            Interface object configured to use given serial parameters.
            NOTE: the instance has not yet tried to contact the hw
            - caller must call 'connect()' before trying to use the Jade.
        """
        if device and JadeTCPImpl.isSupportedDevice(device):
            impl = JadeTCPImpl(device, timeout or DEFAULT_SERIAL_TIMEOUT)
        else:
            impl = JadeSerialImpl(device,
                                  baud or DEFAULT_BAUD_RATE,
                                  timeout or DEFAULT_SERIAL_TIMEOUT)
        return JadeInterface(impl)

    @staticmethod
    def create_ble(device_name=None, serial_number=None,
                   scan_timeout=None, loop=None):
        """
        Create a JadeInterface object using the BLE interface described.
        NOTE: raises JadeError if BLE dependencies not installed.

        Parameters
        ----------
        device_name : str, optional
            The device name of the desired BLE device.
            Underlying implementation will default (to 'Jade')

        serial_number : int, optional
            The serial number of the desired BLE device
            - used to disambiguate multiple beacons with the same 'device name'
            Underlying implementation will connect to the first beacon it scans
            with the matching 'device name'.

        scan_timeout : int, optional
            The timeout when scanning for devices which match the device name/serial number.
            Underlying implementation will default (to 60s)

        loop : optional
            The asynchio event loop to use, if required.
            Underlying implementation will default (to asyncio.get_event_loop())

        Returns
        -------
        JadeInterface
            Interface object configured to use given BLE parameters.
            NOTE: the instance has not yet tried to contact the hw
            - caller must call 'connect()' before trying to use the Jade.

        Raises
        ------
        JadeError if BLE backend not available (ie. BLE dependencies not installed)
        """
        this_module = sys.modules[__name__]
        if not hasattr(this_module, "JadeBleImpl"):
            raise JadeError(1, "BLE support not installed", None)

        impl = JadeBleImpl(device_name or DEFAULT_BLE_DEVICE_NAME,
                           serial_number or DEFAULT_BLE_SERIAL_NUMBER,
                           scan_timeout or DEFAULT_BLE_SCAN_TIMEOUT,
                           loop=loop)
        return JadeInterface(impl)

    def connect(self):
        """
        Try to connect the underlying transport interface (eg. serial, ble, etc.)
        Raises an exception on failure.
        """
        self.impl.connect()

    def disconnect(self, drain=False):
        """
        Disconnect the underlying transport (eg. serial, ble, etc.)

        Parameters
        ----------
        drain : bool, optional
            When true log any/all remaining messages/data, otherwise silently discard.
            NOTE: can prevent disconnection if data is arriving constantly.
            Defaults to False.
        """
        if drain:
            self.drain()

        self.impl.disconnect()

    def drain(self):
        """
        Log any/all outstanding messages/data.
        NOTE: can run indefinitely if data is arriving constantly.
        """
        logger.warning("Draining interface...")
        drained = bytearray()
        finished = False

        while not finished:
            byte_ = self.impl.read(1)
            drained.extend(byte_)
            finished = byte_ == b''

            if finished or byte_ == b'\n' or len(drained) > 256:
                try:
                    device_logger.warning(drained.decode('utf-8'))
                except Exception as e:
                    # Dump the bytes raw and as hex if decoding as utf-8 failed
                    device_logger.warning("Raw:")
                    device_logger.warning(drained)
                    device_logger.warning("----")
                    device_logger.warning("Hex dump:")
                    device_logger.warning(drained.hex())

                # Clear and loop to continue collecting
                drained.clear()

    @staticmethod
    def build_request(input_id, method, params=None):
        """
        Build a request dict from passed parameters

        Parameters
        ----------
        input_id : str
            The id of the request message to construct

        method : str
            rpc method to invoke

        params : dict, optional
            any parameters to pass to the rpc method
            Defaults to None.

        Returns
        -------
        dict
            The request object as a dict
        """
        request = {"method": method, "id": input_id}
        if params is not None:
            request["params"] = params
        return request

    @staticmethod
    def serialise_cbor_request(request):
        """
        Method to format a request dict as a cbor message

        Parameters
        ----------
        request : dict
            The request dict

        Returns
        -------
        bytes
            The request formatted as cbor message bytes
        """
        dump = cbor.dumps(request)
        len_dump = len(dump)
        if 'method' in request and 'ota_data' in request['method']:
            msg = 'Sending ota_data message {} as cbor of size {}'.format(request['id'], len_dump)
            logger.info(msg)
        else:
            logger.info('Sending: {} as cbor of size {}'.format(_hexlify(request), len_dump))
        return dump

    def write(self, bytes_):
        """
        Write bytes over the underlying interface

        Parameters
        ----------
        bytes_ : bytes
            The bytes to write

        Returns
        -------
        int
            The number of bytes written
        """
        logger.debug("Sending: {} bytes".format(len(bytes_)))
        wrote = self.impl.write(bytes_)
        logger.debug("Sent: {} bytes".format(len(bytes_)))
        return wrote

    def write_request(self, request):
        """
        Write a request dict over the underlying interface, formatted as cbor.

        Parameters
        ----------
        request : dict
            The request dict to write
        """
        msg = self.serialise_cbor_request(request)
        written = 0
        while written < len(msg):
            written += self.write(msg[written:])

    def read(self, n):
        """
        Try to read bytes from the underlying interface.

        Returns
        -------
        bytes
            The bytes received
        """
        logger.debug("Reading {} bytes...".format(n))
        bytes_ = self.impl.read(n)
        logger.debug("Received: {} bytes".format(len(bytes_)))
        return bytes_

    def read_cbor_message(self):
        """
        Try to read a single cbor (response) message from the underlying interface.
        Respects the any read timeout.
        If any 'log' messages are received, logs them locally at the nearest corresponding level
        and awaits the next message.  Returns when it receives what appears to be a reply message.

        Returns
        -------
        dict
            The message received, as a dict
        """
        while True:
            # 'self' is sufficiently 'file-like' to act as a load source.
            # Throws EOFError on end of stream/timeout/lost-connection etc.
            message = cbor.load(self)

            if isinstance(message, collections.abc.Mapping):
                # A message response (to a prior request)
                if 'id' in message:
                    logger.info("Received msg: {}".format(_hexlify(message)))
                    return message

                # A log message - handle as normal
                if 'log' in message:
                    response = message['log']
                    log_method = device_logger.error
                    try:
                        response = message['log'].decode("utf-8")
                        log_methods = {
                            'E': device_logger.error,
                            'W': device_logger.warning,
                            'I': device_logger.info,
                            'D': device_logger.debug,
                            'V': device_logger.debug,
                        }
                        if len(response) > 1 and response[1] == ' ':
                            lvl = response[0]
                            log_method = log_methods.get(lvl, device_logger.error)
                    except Exception as e:
                        logger.error('Error processing log message: {}'.format(e))
                    log_method('>> {}'.format(response))
                    continue

            # Unknown/unhandled/unexpected message
            logger.error("Unhandled message received")
            device_logger.error(message)

    def read_response(self, long_timeout=False):
        """
        Try to read a single cbor (response) message from the underlying interface.
        If any 'log' messages are received, logs them locally at the nearest corresponding level
        and awaits the next message.  Returns when it receives what appears to be a reply message.
        If `long_timeout` is false, any read-timeout is respected.  If True, the call will block
        indefinitely awaiting a response message.

        Parameters
        ----------
        long_timeout : bool
            Whether to wait indefinitely for the next (response) message.

        Returns
        -------
        dict
            The message received, as a dict
        """
        while True:
            try:
                return self.read_cbor_message()
            except EOFError as e:
                if not long_timeout:
                    raise

    @staticmethod
    def validate_reply(request, reply):
        """
        Helper to minimally validate a reply message, in the context of a request.
        Asserts if the reply does contain the expected minimal fields.
        """
        assert isinstance(reply, dict) and 'id' in reply
        assert ('result' in reply) != ('error' in reply)
        assert reply['id'] == request['id'] or \
            reply['id'] == '00' and 'error' in reply

    def make_rpc_call(self, request, long_timeout=False):
        """
        Method to send a request over the underlying interface, and await a response.
        The request is minimally validated before it is sent, and the response is similarly
        validated before being returned.
        Any read-timeout is respected unless 'long_timeout' is passed, in which case the call
        blocks indefinitely awaiting a response.

        Parameters
        ----------
        long_timeout : bool
            Whether to wait indefinitely for the response.

        Returns
        -------
        dict
            The (minimally validated) response message received, as a dict
        """
        # Write outgoing request message
        assert isinstance(request, dict)
        assert 'id' in request and len(request['id']) > 0
        assert 'method' in request and len(request['method']) > 0
        assert len(request['id']) < 16 and len(request['method']) < 32
        self.write_request(request)

        # Read and validate incoming message
        reply = self.read_response(long_timeout)
        self.validate_reply(request, reply)

        return reply



# Copyright (C) 2018 Adam Gibson (waxwing)
# Copyright (C) 2018 The Electrum developers
# Distributed under the MIT software license, see the accompanying
# file LICENCE or http://www.opensource.org/licenses/mit-license.php

# Derived from https://gist.github.com/AdamISZ/046d05c156aaeb56cc897f85eecb3eb8

import re
import hashlib
import asyncio
from asyncio import StreamReader, StreamWriter
from functools import cached_property
from typing import NamedTuple, List, Tuple, Mapping, Optional, TYPE_CHECKING, Union, Dict, Set, Sequence

from aiorpcx import NetAddress
import electrum_ecc as ecc

from .crypto import sha256, hmac_oneshot, chacha20_poly1305_encrypt, chacha20_poly1305_decrypt, get_ecdh, privkey_to_pubkey
from .util import ESocksProxy


class LightningPeerConnectionClosed(Exception): pass
class HandshakeFailed(Exception): pass
class ConnStringFormatError(Exception): pass


if TYPE_CHECKING:
    from electrum.network import Network


class HandshakeState(object):
    prologue = b"lightning"
    protocol_name = b"Noise_XK_secp256k1_ChaChaPoly_SHA256"
    handshake_version = b"\x00"

    def __init__(self, responder_pub):
        self.responder_pub = responder_pub
        self.h = sha256(self.protocol_name)
        self.ck = self.h
        self.update(self.prologue)
        self.update(self.responder_pub)

    def update(self, data):
        self.h = sha256(self.h + data)
        return self.h


def get_nonce_bytes(n):
    """BOLT 8 requires the nonce to be 12 bytes, 4 bytes leading
    zeroes and 8 bytes little endian encoded 64 bit integer.
    """
    return b"\x00"*4 + n.to_bytes(8, 'little')


def aead_encrypt(key: bytes, nonce: int, associated_data: bytes, data: bytes) -> bytes:
    nonce_bytes = get_nonce_bytes(nonce)
    return chacha20_poly1305_encrypt(key=key,
                                     nonce=nonce_bytes,
                                     associated_data=associated_data,
                                     data=data)


def aead_decrypt(key: bytes, nonce: int, associated_data: bytes, data: bytes) -> bytes:
    nonce_bytes = get_nonce_bytes(nonce)
    return chacha20_poly1305_decrypt(key=key,
                                     nonce=nonce_bytes,
                                     associated_data=associated_data,
                                     data=data)


def get_bolt8_hkdf(salt, ikm):
    """RFC5869 HKDF instantiated in the specific form
    used in Lightning BOLT 8:
    Extract and expand to 64 bytes using HMAC-SHA256,
    with info field set to a zero length string as per BOLT8
    Return as two 32 byte fields.
    """
    #Extract
    prk = hmac_oneshot(salt, msg=ikm, digest=hashlib.sha256)
    assert len(prk) == 32
    #Expand
    info = b""
    T0 = b""
    T1 = hmac_oneshot(prk, T0 + info + b"\x01", digest=hashlib.sha256)
    T2 = hmac_oneshot(prk, T1 + info + b"\x02", digest=hashlib.sha256)
    assert len(T1 + T2) == 64
    return T1, T2


def act1_initiator_message(hs, epriv, epub):
    ss = get_ecdh(epriv, hs.responder_pub)
    ck2, temp_k1 = get_bolt8_hkdf(hs.ck, ss)
    hs.ck = ck2
    c = aead_encrypt(temp_k1, 0, hs.update(epub), b"")
    #for next step if we do it
    hs.update(c)
    msg = hs.handshake_version + epub + c
    assert len(msg) == 50
    return msg, temp_k1


def create_ephemeral_key() -> (bytes, bytes):
    privkey = ecc.ECPrivkey.generate_random_key()
    return privkey.get_secret_bytes(), privkey.get_public_key_bytes()


def split_host_port(host_port: str) -> Tuple[str, str]: # port returned as string
    ipv6  = re.compile(r'\[(?P<host>[:0-9a-f]+)\](?P<port>:\d+)?$')
    other = re.compile(r'(?P<host>[^:]+)(?P<port>:\d+)?$')
    m = ipv6.match(host_port)
    if not m:
        m = other.match(host_port)
    if not m:
        raise ConnStringFormatError('Connection strings must be in <node_pubkey>@<host>:<port> format')
    host = m.group('host')
    if m.group('port'):
        port = m.group('port')[1:]
    else:
        port = '9735'
    try:
        int(port)
    except ValueError:
        raise ConnStringFormatError('Port number must be decimal')
    return host, port


def extract_nodeid(connect_contents: str) -> Tuple[bytes, Optional[str]]:
    """Takes a connection-string-like str, and returns a tuple (node_id, rest),
    where rest is typically a host (with maybe port). Examples:
    - extract_nodeid(pubkey@host:port) == (pubkey, host:port)
    - extract_nodeid(pubkey@host) == (pubkey, host)
    - extract_nodeid(pubkey) == (pubkey, None)
    Can raise ConnStringFormatError.
    """
    rest = None
    try:
        # connection string?
        nodeid_hex, rest = connect_contents.split("@", 1)
    except ValueError:
        # node id as hex?
        nodeid_hex = connect_contents
    if rest == '':
        raise ConnStringFormatError('At least a hostname must be supplied after the at symbol.')
    try:
        node_id = bytes.fromhex(nodeid_hex)
        if len(node_id) != 33:
            raise Exception()
    except Exception:
        raise ConnStringFormatError('Invalid node ID, must be 33 bytes and hexadecimal')
    return node_id, rest


class LNPeerAddr:
    # note: while not programmatically enforced, this class is meant to be *immutable*

    def __init__(self, host: str, port: int, pubkey: bytes):
        assert isinstance(host, str), repr(host)
        assert isinstance(port, int), repr(port)
        assert isinstance(pubkey, bytes), repr(pubkey)
        try:
            net_addr = NetAddress(host, port)  # this validates host and port
        except Exception as e:
            raise ValueError(f"cannot construct LNPeerAddr: invalid host or port (host={host}, port={port})") from e
        # note: not validating pubkey as it would be too expensive:
        # if not ECPubkey.is_pubkey_bytes(pubkey): raise ValueError()
        self.host = host
        self.port = port
        self.pubkey = pubkey
        self._net_addr = net_addr

    def __str__(self):
        return '{}@{}'.format(self.pubkey.hex(), self.net_addr_str())

    @classmethod
    def from_str(cls, s):
        node_id, rest = extract_nodeid(s)
        host, port = split_host_port(rest)
        return LNPeerAddr(host, int(port), node_id)

    def __repr__(self):
        return f'<LNPeerAddr host={self.host} port={self.port} pubkey={self.pubkey.hex()}>'

    def net_addr(self) -> NetAddress:
        return self._net_addr

    def net_addr_str(self) -> str:
        return str(self._net_addr)

    def is_onion(self) -> bool:
        return self.host.endswith('.onion')

    def __eq__(self, other):
        if not isinstance(other, LNPeerAddr):
            return False
        return (self.host == other.host
                and self.port == other.port
                and self.pubkey == other.pubkey)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.host, self.port, self.pubkey))


class LNTransportBase:
    reader: StreamReader
    writer: StreamWriter
    privkey: bytes
    peer_addr: Optional[LNPeerAddr] = None

    def __init__(self):
        self.drain_write_lock = asyncio.Lock()

    def name(self) -> str:
        pubkey = self.remote_pubkey()
        pubkey_hex = pubkey.hex() if pubkey else pubkey
        return f"{pubkey_hex[:10]}-{self._id_hash[:8]}"

    @cached_property
    def _id_hash(self) -> str:
        id_int = id(self)
        id_bytes = id_int.to_bytes((id_int.bit_length() + 7) // 8, byteorder='big')
        return sha256(id_bytes).hex()

    def send_bytes(self, msg: bytes) -> None:
        l = len(msg).to_bytes(2, 'big')
        lc = aead_encrypt(self.sk, self.sn(), b'', l)
        c = aead_encrypt(self.sk, self.sn(), b'', msg)
        assert len(lc) == 18
        assert len(c) == len(msg) + 16
        self.writer.write(lc+c)

    async def send_bytes_and_drain(self, msg: bytes) -> None:
        """Should be used when possible (in async scope), to avoid memory exhaustion."""
        async with self.drain_write_lock:
            self.send_bytes(msg)
            try:
                await self.writer.drain()
            except ConnectionError as e:
                raise LightningPeerConnectionClosed() from e

    async def read_messages(self):
        buffer = bytearray()
        while True:
            rn_l, rk_l = self.rn()
            rn_m, rk_m = self.rn()
            while True:
                if len(buffer) >= 18:
                    lc = bytes(buffer[:18])
                    l = aead_decrypt(rk_l, rn_l, b'', lc)
                    length = int.from_bytes(l, 'big')
                    offset = 18 + length + 16
                    if len(buffer) >= offset:
                        c = bytes(buffer[18:offset])
                        del buffer[:offset]  # much faster than: buffer=buffer[offset:]
                        msg = aead_decrypt(rk_m, rn_m, b'', c)
                        yield msg
                        break
                try:
                    s = await self.reader.read(2**10)
                except Exception:
                    s = None
                if not s:
                    raise LightningPeerConnectionClosed()
                buffer += s

    def rn(self):
        o = self._rn, self.rk
        self._rn += 1
        if self._rn == 1000:
            self.r_ck, self.rk = get_bolt8_hkdf(self.r_ck, self.rk)
            self._rn = 0
        return o

    def sn(self):
        o = self._sn
        self._sn += 1
        if self._sn == 1000:
            self.s_ck, self.sk = get_bolt8_hkdf(self.s_ck, self.sk)
            self._sn = 0
        return o

    def init_counters(self, ck):
        # init counters
        self._sn = 0
        self._rn = 0
        self.r_ck = ck
        self.s_ck = ck

    def close(self):
        self.writer.close()

    def remote_pubkey(self) -> Optional[bytes]:
        raise NotImplementedError()


class LNResponderTransport(LNTransportBase):
    """Transport initiated by remote party."""

    def __init__(self, privkey: bytes, reader: StreamReader, writer: StreamWriter):
        LNTransportBase.__init__(self)
        self.reader = reader
        self.writer = writer
        self.privkey = privkey
        self._pubkey = None  # remote pubkey

    def name(self) -> str:
        return f"{super().name()}(in)"

    async def handshake(self, **kwargs):
        hs = HandshakeState(privkey_to_pubkey(self.privkey))
        act1 = b''
        while len(act1) < 50:
            buf = await self.reader.read(50 - len(act1))
            if not buf:
                raise HandshakeFailed('responder disconnected')
            act1 += buf
        if len(act1) != 50:
            raise HandshakeFailed('responder: short act 1 read, length is ' + str(len(act1)))
        if bytes([act1[0]]) != HandshakeState.handshake_version:
            raise HandshakeFailed('responder: bad handshake version in act 1')
        c = act1[-16:]
        re = act1[1:34]
        h = hs.update(re)
        ss = get_ecdh(self.privkey, re)
        ck, temp_k1 = get_bolt8_hkdf(sha256(HandshakeState.protocol_name), ss)
        _p = aead_decrypt(temp_k1, 0, h, c)
        hs.update(c)

        # act 2
        if 'epriv' not in kwargs:
            epriv, epub = create_ephemeral_key()
        else:
            epriv = kwargs['epriv']
            epub = ecc.ECPrivkey(epriv).get_public_key_bytes()
        hs.ck = ck
        hs.responder_pub = re

        msg, temp_k2 = act1_initiator_message(hs, epriv, epub)
        self.writer.write(msg)

        # act 3
        act3 = b''
        while len(act3) < 66:
            buf = await self.reader.read(66 - len(act3))
            if not buf:
                raise HandshakeFailed('responder disconnected')
            act3 += buf
        if len(act3) != 66:
            raise HandshakeFailed('responder: short act 3 read, length is ' + str(len(act3)))
        if bytes([act3[0]]) != HandshakeState.handshake_version:
            raise HandshakeFailed('responder: bad handshake version in act 3')
        c = act3[1:50]
        t = act3[-16:]
        rs = aead_decrypt(temp_k2, 1, hs.h, c)
        ss = get_ecdh(epriv, rs)
        ck, temp_k3 = get_bolt8_hkdf(hs.ck, ss)
        _p = aead_decrypt(temp_k3, 0, hs.update(c), t)
        self.rk, self.sk = get_bolt8_hkdf(ck, b'')
        self.init_counters(ck)
        self._pubkey = rs
        return rs

    def remote_pubkey(self) -> Optional[bytes]:
        return self._pubkey


class LNTransport(LNTransportBase):
    """Transport initiated by local party."""

    def __init__(self, privkey: bytes, peer_addr: LNPeerAddr, *,
                 e_proxy: Optional['ESocksProxy']):
        LNTransportBase.__init__(self)
        assert type(privkey) is bytes and len(privkey) == 32
        self.privkey = privkey
        self.peer_addr = peer_addr
        self.e_proxy = e_proxy

    async def handshake(self):
        if not self.e_proxy:
            self.reader, self.writer = await asyncio.open_connection(self.peer_addr.host, self.peer_addr.port)
        else:
            self.reader, self.writer = await self.e_proxy.open_connection(self.peer_addr.host, self.peer_addr.port)
        hs = HandshakeState(self.peer_addr.pubkey)
        # Get a new ephemeral key
        epriv, epub = create_ephemeral_key()

        msg, _temp_k1 = act1_initiator_message(hs, epriv, epub)
        # act 1
        self.writer.write(msg)
        rspns = await self.reader.read(2**10)
        if len(rspns) != 50:
            raise HandshakeFailed(f"Lightning handshake act 1 response has bad length, "
                                  f"are you sure this is the right pubkey? {self.peer_addr}")
        hver, alice_epub, tag = rspns[0], rspns[1:34], rspns[34:]
        if bytes([hver]) != hs.handshake_version:
            raise HandshakeFailed("unexpected handshake version: {}".format(hver))
        # act 2
        hs.update(alice_epub)
        ss = get_ecdh(epriv, alice_epub)
        ck, temp_k2 = get_bolt8_hkdf(hs.ck, ss)
        hs.ck = ck
        p = aead_decrypt(temp_k2, 0, hs.h, tag)
        hs.update(tag)
        # act 3
        my_pubkey = privkey_to_pubkey(self.privkey)
        c = aead_encrypt(temp_k2, 1, hs.h, my_pubkey)
        hs.update(c)
        ss = get_ecdh(self.privkey[:32], alice_epub)
        ck, temp_k3 = get_bolt8_hkdf(hs.ck, ss)
        hs.ck = ck
        t = aead_encrypt(temp_k3, 0, hs.h, b'')
        msg = hs.handshake_version + c + t
        self.writer.write(msg)
        self.sk, self.rk = get_bolt8_hkdf(hs.ck, b'')
        self.init_counters(ck)

    def remote_pubkey(self) -> Optional[bytes]:
        return self.peer_addr.pubkey



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2012 thomasv@gitorious
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from functools import partial
from typing import Optional, TYPE_CHECKING, Union

from PyQt6.QtCore import Qt, QTimer, QSize, QStringListModel
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QFontMetrics, QFont, QContextMenuEvent
from PyQt6.QtWidgets import QTextEdit, QWidget, QLineEdit, QStackedLayout, QCompleter

from electrum.payment_identifier import PaymentIdentifier
from electrum.logging import Logger
from electrum.util import EventListener, event_listener

from . import util
from .util import MONOSPACE_FONT, GenericInputHandler, ColorScheme, add_input_actions_to_context_menu

if TYPE_CHECKING:
    from .send_tab import SendTab


frozen_style = "QWidget {border:none;}"
normal_style = "QPlainTextEdit { }"


class InvalidPaymentIdentifier(Exception):
    pass


class ResizingTextEdit(QTextEdit):

    textReallyChanged = pyqtSignal()
    resized = pyqtSignal()

    def __init__(self):
        QTextEdit.__init__(self)
        self._text = ''
        self.setAcceptRichText(False)
        self.textChanged.connect(self.on_text_changed)
        document = self.document()
        fontMetrics = QFontMetrics(document.defaultFont())
        self.fontSpacing = fontMetrics.lineSpacing()
        margins = self.contentsMargins()
        documentMargin = document.documentMargin()
        self.verticalMargins = margins.top() + margins.bottom()
        self.verticalMargins += self.frameWidth() * 2
        self.verticalMargins += documentMargin * 2
        self.heightMin = self.fontSpacing + self.verticalMargins
        self.heightMax = (self.fontSpacing * 10) + self.verticalMargins
        self.update_size()

    def on_text_changed(self):
        # QTextEdit emits spurious textChanged events
        if self.toPlainText() != self._text:
            self._text = self.toPlainText()
            self.textReallyChanged.emit()
            self.update_size()

    def update_size(self):
        docLineCount = self.document().lineCount()
        docHeight = max(3, docLineCount) * self.fontSpacing
        h = docHeight + self.verticalMargins
        h = min(max(h, self.heightMin), self.heightMax)
        self.setMinimumHeight(int(h))
        self.setMaximumHeight(int(h))
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.verticalScrollBar().setHidden(docHeight + self.verticalMargins < self.heightMax)
        self.setLineWrapMode(QTextEdit.LineWrapMode.WidgetWidth)
        self.resized.emit()

    def sizeHint(self) -> QSize:
        return QSize(0, self.minimumHeight())


class PayToEdit(QWidget, Logger, GenericInputHandler, EventListener):
    paymentIdentifierChanged = pyqtSignal()
    textChanged = pyqtSignal()

    def __init__(self, send_tab: 'SendTab'):
        QWidget.__init__(self, parent=send_tab)
        Logger.__init__(self)
        GenericInputHandler.__init__(self)

        self._text = ''
        self._layout = QStackedLayout()
        self.setLayout(self._layout)

        self.send_tab = send_tab

        def text_edit_changed():
            text = self.text_edit.toPlainText()
            if self._text != text:
                # sync and emit
                self._text = text
                self.line_edit.setText(text)
                self.textChanged.emit()

        def text_edit_resized():
            self.update_height()

        def line_edit_changed():
            text = self.line_edit.text()
            if self._text != text:
                # sync and emit
                self._text = text
                self.text_edit.setPlainText(text)
                self.textChanged.emit()

        self.line_edit = QLineEdit()
        self.line_edit.textChanged.connect(line_edit_changed)
        self.text_edit = ResizingTextEdit()
        self.text_edit.setTabChangesFocus(True)
        self.text_edit.textReallyChanged.connect(text_edit_changed)
        self.text_edit.resized.connect(text_edit_resized)

        def on_completed(item: str):
            text = self._completer_contacts[1][self._completer_contacts[0].index(item)]
            self.try_payment_identifier(text)
            self.completer.popup().hide()

        self.completer = QCompleter()
        self.completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self.completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.completer.setFilterMode(Qt.MatchFlag.MatchContains)
        self.completer.activated.connect(on_completed)

        self.update_completer()

        self.line_edit.setCompleter(self.completer)

        self.textChanged.connect(self._handle_text_change)

        self._layout.addWidget(self.line_edit)
        self._layout.addWidget(self.text_edit)

        self.multiline = False

        self._is_paytomany = False
        self.line_edit.setFont(QFont(MONOSPACE_FONT))
        self.text_edit.setFont(QFont(MONOSPACE_FONT))
        self.send_tab = send_tab
        self.config = send_tab.config

        # button handlers
        self.on_qr_from_camera_input_btn = partial(
            self.input_qr_from_camera,
            config=self.config,
            allow_multi=False,
            show_error=self.send_tab.show_error,
            setText=self.try_payment_identifier,
            parent=self.send_tab.window,
        )
        self.on_qr_from_screenshot_input_btn = partial(
            self.input_qr_from_screenshot,
            allow_multi=False,
            show_error=self.send_tab.show_error,
            setText=self.try_payment_identifier,
        )
        self.on_qr_from_file_input_btn = partial(
            self.input_qr_from_file,
            allow_multi=False,
            config=self.config,
            show_error=self.send_tab.show_error,
            setText=self.try_payment_identifier,
        )
        self.on_input_file = partial(
            self.input_file,
            config=self.config,
            show_error=self.send_tab.show_error,
            setText=self.try_payment_identifier,
        )

        self.text_edit.contextMenuEvent = partial(self.custom_context_menu_event, tl_edit=self.text_edit)
        self.line_edit.contextMenuEvent = partial(self.custom_context_menu_event, tl_edit=self.line_edit)

        self.edit_timer = QTimer(self)
        self.edit_timer.setSingleShot(True)
        self.edit_timer.setInterval(1000)
        self.edit_timer.timeout.connect(self._on_edit_timer)

        self.payment_identifier = None  # type: Optional[PaymentIdentifier]

        self.register_callbacks()
        self.destroyed.connect(lambda: self.unregister_callbacks())

    def custom_context_menu_event(self, e: 'QContextMenuEvent', *, tl_edit: Union[QTextEdit, QLineEdit]) -> None:
        m = tl_edit.createStandardContextMenu()
        m.addSeparator()
        add_input_actions_to_context_menu(self, m)
        m.exec(e.globalPos())

    @event_listener
    def on_event_contacts_updated(self):
        self.update_completer()

    def update_completer(self):
        self._completer_contacts = [], []
        for k, v in self.send_tab.wallet.contacts.items():
            self._completer_contacts[0].append(f'{v[1]} <{k}>')
            self._completer_contacts[1].append(k)

        self.completer.setModel(QStringListModel(self._completer_contacts[0]))

    @property
    def multiline(self):
        return self._multiline

    @multiline.setter
    def multiline(self, b: bool) -> None:
        if b is None:
            return
        self._multiline = b
        self._layout.setCurrentWidget(self.text_edit if b else self.line_edit)
        self.update_height()

    def update_height(self) -> None:
        h = self._layout.currentWidget().sizeHint().height()
        self.setMaximumHeight(h)

    def setText(self, text: str) -> None:
        if self._text != text:
            self.line_edit.setText(text)
            self.text_edit.setText(text)

    def setFocus(self, reason=Qt.FocusReason.OtherFocusReason) -> None:
        if self.multiline:
            self.text_edit.setFocus(reason)
        else:
            self.line_edit.setFocus(reason)

    def setToolTip(self, tt: str) -> None:
        self.line_edit.setToolTip(tt)
        self.text_edit.setToolTip(tt)

    def try_payment_identifier(self, text) -> None:
        '''set payment identifier only if valid, else exception'''
        pi = PaymentIdentifier(self.send_tab.wallet, text)
        if not pi.is_valid():
            raise InvalidPaymentIdentifier('Invalid payment identifier')
        self.set_payment_identifier(text)

    def set_payment_identifier(self, text) -> None:
        if self.payment_identifier and self.payment_identifier.text == text.strip():
            # no change.
            return

        self.payment_identifier = PaymentIdentifier(self.send_tab.wallet, text)

        # toggle to multiline if payment identifier is a multiline
        if self.payment_identifier.is_multiline() and not self._is_paytomany:
            self.set_paytomany(True)

        # if payment identifier gets set externally, we want to update the edit control
        # Note: this triggers the change handler, but we shortcut if it's the same payment identifier
        self.setText(text)

        self.paymentIdentifierChanged.emit()

    def set_paytomany(self, b):
        self._is_paytomany = b
        self.multiline = b
        self.send_tab.paytomany_menu.setChecked(b)

    def toggle_paytomany(self) -> None:
        self.set_paytomany(not self._is_paytomany)

    def is_paytomany(self):
        return self._is_paytomany

    def setReadOnly(self, b: bool) -> None:
        self.line_edit.setReadOnly(b)
        self.text_edit.setReadOnly(b)

    def isReadOnly(self):
        return self.line_edit.isReadOnly()

    def setStyleSheet(self, stylesheet: str) -> None:
        self.line_edit.setStyleSheet(stylesheet)
        self.text_edit.setStyleSheet(stylesheet)

    def setFrozen(self, b) -> None:
        self.setReadOnly(b)
        self.setStyleSheet(ColorScheme.LIGHTBLUE.as_stylesheet(True) if b else '')

    def isFrozen(self):
        return self.isReadOnly()

    def do_clear(self) -> None:
        self.set_paytomany(False)
        self.setText('')
        self.setToolTip('')
        self.payment_identifier = None

    def setGreen(self) -> None:
        self.setStyleSheet(util.ColorScheme.GREEN.as_stylesheet(True))

    def setExpired(self) -> None:
        self.setStyleSheet(util.ColorScheme.RED.as_stylesheet(True))

    def _handle_text_change(self) -> None:
        if self.isFrozen():
            # if editor is frozen, we ignore text changes as they might not be a payment identifier
            # but a user friendly representation.
            return

        # pushback timer if timer active or PI needs resolving
        pi = PaymentIdentifier(self.send_tab.wallet, self._text)
        if not pi.is_valid() or pi.need_resolve() or self.edit_timer.isActive():
            self.edit_timer.start()
        else:
            self.set_payment_identifier(self._text)

    def _on_edit_timer(self) -> None:
        if not self.isFrozen():
            self.set_payment_identifier(self._text)



#!/usr/bin/env python3
import json
import asyncio

from electrum.simple_config import SimpleConfig
from electrum.network import filter_version, Network
from electrum.util import create_and_start_event_loop, log_exceptions
from electrum import constants

# testnet?
config = SimpleConfig({'testnet': False})
config.get_selected_chain().set_as_network()

loop, stopping_fut, loop_thread = create_and_start_event_loop()
network = Network(config)
network.start()

@log_exceptions
async def f():
    try:
        peers = await network.get_peers()
        peers = filter_version(peers)
        print(json.dumps(peers, sort_keys=True, indent=4))
    finally:
        stopping_fut.set_result(1)

asyncio.run_coroutine_threadsafe(f(), loop)
while loop_thread.is_alive():
    loop_thread.join(1)



from electrum.plugin import hook
from electrum.hw_wallet import CmdLineHandler
from .keepkey import KeepKeyPlugin

class Plugin(KeepKeyPlugin):
    handler = CmdLineHandler()
    @hook
    def init_keystore(self, keystore):
        if not isinstance(keystore, self.keystore_class):
            return
        keystore.handler = self.handler

    def create_handler(self, window):
        return self.handler



import random
import math
from typing import List, Tuple, Dict, NamedTuple

from .lnutil import NoPathFound

PART_PENALTY = 1.0  # 1.0 results in avoiding splits
MIN_PART_SIZE_MSAT = 10_000_000  # we don't want to split indefinitely
EXHAUST_DECAY_FRACTION = 10  # fraction of the local balance that should be reserved if possible
RELATIVE_SPLIT_SPREAD = 0.3  # deviation from the mean when splitting amounts into parts

# these parameters affect the computational work in the probabilistic algorithm
CANDIDATES_PER_LEVEL = 20
MAX_PARTS = 5  # maximum number of parts for splitting


# maps a channel (channel_id, node_id) to the funds it has available
ChannelsFundsInfo = Dict[Tuple[bytes, bytes], Tuple[int, int]]


class SplitConfig(dict, Dict[Tuple[bytes, bytes], List[int]]):
    """maps a channel (channel_id, node_id) to a list of amounts"""
    def number_parts(self) -> int:
        return sum([len(v) for v in self.values() if sum(v)])

    def number_nonzero_channels(self) -> int:
        return len([v for v in self.values() if sum(v)])

    def number_nonzero_nodes(self) -> int:
        # using a set comprehension
        return len({nodeid for (_, nodeid), amounts in self.items() if sum(amounts)})

    def total_config_amount(self) -> int:
        return sum([sum(c) for c in self.values()])

    def is_any_amount_smaller_than_min_part_size(self) -> bool:
        smaller = False
        for amounts in self.values():
            if any([amount < MIN_PART_SIZE_MSAT for amount in amounts]):
                smaller |= True
        return smaller


class SplitConfigRating(NamedTuple):
    config: SplitConfig
    rating: float


def split_amount_normal(total_amount: int, num_parts: int) -> List[int]:
    """Splits an amount into about `num_parts` parts, where the parts are split
    randomly (normally distributed around amount/num_parts with certain spread)."""
    parts = []
    avg_amount = total_amount / num_parts
    # roughly reach total_amount
    while total_amount - sum(parts) > avg_amount:
        amount_to_add = int(abs(random.gauss(avg_amount, RELATIVE_SPLIT_SPREAD * avg_amount)))
        if sum(parts) + amount_to_add < total_amount:
            parts.append(amount_to_add)
    # add what's missing
    parts.append(total_amount - sum(parts))
    return parts


def remove_duplicates(configs: List[SplitConfig]) -> List[SplitConfig]:
    unique_configs = set()
    for config in configs:
        # sort keys and values
        config_sorted_values = {k: sorted(v) for k, v in config.items()}
        config_sorted_keys = {k: config_sorted_values[k] for k in sorted(config_sorted_values.keys())}
        hashable_config = tuple((c, tuple(sorted(config[c]))) for c in config_sorted_keys)
        unique_configs.add(hashable_config)
    unique_configs = [SplitConfig({c[0]: list(c[1]) for c in config}) for config in unique_configs]
    return unique_configs


def remove_multiple_nodes(configs: List[SplitConfig]) -> List[SplitConfig]:
    return [config for config in configs if config.number_nonzero_nodes() == 1]


def remove_single_part_configs(configs: List[SplitConfig]) -> List[SplitConfig]:
    return [config for config in configs if config.number_parts() != 1]


def remove_single_channel_splits(configs: List[SplitConfig]) -> List[SplitConfig]:
    return [
        config for config in configs
        if all(len(channel_splits) <= 1 for channel_splits in config.values())
    ]

def rate_config(
        config: SplitConfig,
        channels_with_funds: ChannelsFundsInfo) -> float:
    """Defines an objective function to rate a configuration.

    We calculate the normalized L2 norm for a configuration and
    add a part penalty for each nonzero amount. The consequence is that
    amounts that are equally distributed and have less parts are rated
    lowest (best). A penalty depending on the total amount sent over a channel
    counteracts channel exhaustion."""
    rating = 0
    total_amount = config.total_config_amount()

    for channel, amounts in config.items():
        funds, slots = channels_with_funds[channel]
        if amounts:
            for amount in amounts:
                rating += amount * amount / (total_amount * total_amount)  # penalty to favor equal distribution of amounts
                rating += PART_PENALTY * PART_PENALTY  # penalty for each part
            decay = funds / EXHAUST_DECAY_FRACTION
            rating += math.exp((sum(amounts) - funds) / decay)  # penalty for channel exhaustion
    return rating


def suggest_splits(
        amount_msat: int,
        channels_with_funds: ChannelsFundsInfo,
        exclude_single_part_payments=False,
        exclude_multinode_payments=False,
        exclude_single_channel_splits=False
) -> List[SplitConfigRating]:
    """Breaks amount_msat into smaller pieces and distributes them over the
    channels according to the funds they can send.

    Individual channels may be assigned multiple parts. The split configurations
    are returned in sorted order, from best to worst rating.

    Single part payments can be excluded, since they represent legacy payments.
    Split configurations that send via multiple nodes can be excluded as well.
    """

    configs = []
    channel_keys = list(channels_with_funds.keys())

    # generate multiple configurations to get more configurations (there is randomness in this loop)
    for _ in range(CANDIDATES_PER_LEVEL):
        # we want to have configurations with no splitting to many splittings
        for target_parts in range(1, MAX_PARTS):
            config = SplitConfig()
            # randomly split amount into target_parts chunks
            split_amounts = split_amount_normal(amount_msat, target_parts)
            # randomly distribute amounts over channels
            for amount in split_amounts:
                random.shuffle(channel_keys)
                # we check each channel and try to put the funds inside, break if we succeed
                for c in channel_keys:
                    amounts = config.get(c, [])
                    channel_funds, channel_slots = channels_with_funds[c]
                    if sum(amounts) + amount <= channel_funds and len(amounts) < channel_slots:
                        config.setdefault(c, []).append(amount)
                        break
                # if we don't succeed to put the amount anywhere,
                # we try to fill up channels and put the rest somewhere else
                else:
                    distribute_amount = amount
                    for c in channel_keys:
                        channel_funds, channel_slots = channels_with_funds[c]
                        amounts = config.get(c, [])
                        slots_left = channel_slots - len(amounts)
                        if slots_left == 0:
                            # no slot left in that channel
                            continue
                        funds_left = channel_funds - sum(amounts)
                        # it would be good to not fill the full channel if possible
                        add_amount = min(funds_left, distribute_amount)
                        if add_amount:
                            config.setdefault(c, []).append(add_amount)
                            distribute_amount -= add_amount
                        if distribute_amount == 0:
                            break
            if config.total_config_amount() != amount_msat:
                continue
            if target_parts > 1 and config.is_any_amount_smaller_than_min_part_size():
                if target_parts == 2:
                    # if there are already too small parts at the first split excluding single
                    # part payments may return only few configurations, this will allow single part
                    # payments for more payments, if they are too small to split
                    exclude_single_part_payments = False
                continue
            assert config.total_config_amount() == amount_msat
            configs.append(config)
        if not configs:
            raise NoPathFound('Cannot distribute payment over channels.')

    configs = remove_duplicates(configs)

    # we only take configurations that send via a single node (but there can be multiple parts)
    if exclude_multinode_payments:
        configs = remove_multiple_nodes(configs)

    if exclude_single_part_payments:
        configs = remove_single_part_configs(configs)

    if exclude_single_channel_splits:
        configs = remove_single_channel_splits(configs)

    rated_configs = [SplitConfigRating(
        config=c,
        rating=rate_config(c, channels_with_funds)
    ) for c in configs]
    rated_configs.sort(key=lambda x: x.rating)

    return rated_configs



# -*- coding: utf-8 -*-

from decimal import Decimal
from typing import Union

from PyQt6.QtCore import pyqtSignal, Qt, QSize
from PyQt6.QtGui import QPainter
from PyQt6.QtWidgets import (QLineEdit, QStyle, QStyleOptionFrame, QSizePolicy)

from .util import char_width_in_lineedit, ColorScheme

from electrum.util import (format_satoshis_plain, decimal_point_to_base_unit_name,
                           FEERATE_PRECISION, quantize_feerate, DECIMAL_POINT, UI_UNIT_NAME_FEERATE_SAT_PER_VBYTE)
from electrum.bitcoin import COIN, TOTAL_COIN_SUPPLY_LIMIT_IN_BTC

_NOT_GIVEN = object()  # sentinel value


class FreezableLineEdit(QLineEdit):
    frozen = pyqtSignal()

    def setFrozen(self, b):
        self.setReadOnly(b)
        self.setStyleSheet(ColorScheme.LIGHTBLUE.as_stylesheet(True) if b else '')
        self.frozen.emit()

    def isFrozen(self):
        return self.isReadOnly()


class SizedFreezableLineEdit(FreezableLineEdit):

    def __init__(self, *, width: int, parent=None):
        super().__init__(parent)
        self._width = width
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        self.setMaximumWidth(width)

    def sizeHint(self) -> QSize:
        sh = super().sizeHint()
        return QSize(self._width, sh.height())


class AmountEdit(SizedFreezableLineEdit):
    shortcut = pyqtSignal()

    def __init__(self, base_unit, is_int=False, parent=None, *, max_amount=None):
        # This seems sufficient for hundred-BTC amounts with 8 decimals
        width = 16 * char_width_in_lineedit()
        super().__init__(width=width, parent=parent)
        self.base_unit = base_unit
        self.textChanged.connect(self.numbify)
        self.is_int = is_int
        self.is_shortcut = False
        self.extra_precision = 0
        self.max_amount = max_amount

    def decimal_point(self):
        return 8

    def max_precision(self):
        return self.decimal_point() + self.extra_precision

    def numbify(self):
        text = self.text().strip()
        if text == '!':
            self.shortcut.emit()
            return
        pos = self.cursorPosition()
        chars = '0123456789'
        if not self.is_int: chars += DECIMAL_POINT
        s = ''.join([i for i in text if i in chars])
        if not self.is_int:
            if DECIMAL_POINT in s:
                p = s.find(DECIMAL_POINT)
                s = s.replace(DECIMAL_POINT, '')
                s = s[:p] + DECIMAL_POINT + s[p:p+self.max_precision()]
        if self.max_amount:
            if (amt := self._get_amount_from_text(s)) and amt >= self.max_amount:
                s = self._get_text_from_amount(self.max_amount)
        self.setText(s)
        # setText sets Modified to False.  Instead we want to remember
        # if updates were because of user modification.
        self.setModified(self.hasFocus())
        self.setCursorPosition(pos)

    def paintEvent(self, event):
        QLineEdit.paintEvent(self, event)
        if self.base_unit:
            panel = QStyleOptionFrame()
            self.initStyleOption(panel)
            textRect = self.style().subElementRect(QStyle.SubElement.SE_LineEditContents, panel, self)
            textRect.adjust(2, 0, -10, 0)
            painter = QPainter(self)
            painter.setPen(ColorScheme.GRAY.as_color())
            painter.drawText(textRect, int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter), self.base_unit())

    def _get_amount_from_text(self, text: str) -> Union[None, Decimal, int]:
        try:
            text = text.replace(DECIMAL_POINT, '.')
            return (int if self.is_int else Decimal)(text)
        except Exception:
            return None

    def get_amount(self) -> Union[None, Decimal, int]:
        amt = self._get_amount_from_text(str(self.text()))
        if self.max_amount and amt and amt >= self.max_amount:
            return self.max_amount
        return amt

    def _get_text_from_amount(self, amount) -> str:
        return "%d" % amount

    def setAmount(self, amount):
        text = self._get_text_from_amount(amount)
        self.setText(text)


class BTCAmountEdit(AmountEdit):

    def __init__(self, decimal_point, is_int=False, parent=None, *, max_amount=_NOT_GIVEN):
        if max_amount is _NOT_GIVEN:
            max_amount = TOTAL_COIN_SUPPLY_LIMIT_IN_BTC * COIN
        AmountEdit.__init__(self, self._base_unit, is_int, parent, max_amount=max_amount)
        self.decimal_point = decimal_point

    def _base_unit(self):
        return decimal_point_to_base_unit_name(self.decimal_point())

    def _get_amount_from_text(self, text):
        # returns amt in satoshis
        try:
            text = text.replace(DECIMAL_POINT, '.')
            x = Decimal(text)
        except Exception:
            return None
        # scale it to max allowed precision, make it an int
        power = pow(10, self.max_precision())
        max_prec_amount = int(power * x)
        # if the max precision is simply what unit conversion allows, just return
        if self.max_precision() == self.decimal_point():
            return max_prec_amount
        # otherwise, scale it back to the expected unit
        amount = Decimal(max_prec_amount) / pow(10, self.max_precision()-self.decimal_point())
        return Decimal(amount) if not self.is_int else int(amount)

    def _get_text_from_amount(self, amount_sat):
        text = format_satoshis_plain(amount_sat, decimal_point=self.decimal_point())
        text = text.replace('.', DECIMAL_POINT)
        return text

    def setAmount(self, amount_sat):
        if amount_sat is None:
            self.setText(" ")  # Space forces repaint in case units changed
        else:
            text = self._get_text_from_amount(amount_sat)
            self.setText(text)
        self.setFrozen(self.isFrozen()) # re-apply styling, as it is nuked by setText (?)
        self.repaint()  # macOS hack for #6269


class FeerateEdit(BTCAmountEdit):

    def __init__(self, decimal_point, is_int=False, parent=None, *, max_amount=_NOT_GIVEN):
        super().__init__(decimal_point, is_int, parent, max_amount=max_amount)
        self.extra_precision = FEERATE_PRECISION

    def _base_unit(self):
        return UI_UNIT_NAME_FEERATE_SAT_PER_VBYTE

    def _get_amount_from_text(self, text):
        sat_per_byte_amount = super()._get_amount_from_text(text)
        return quantize_feerate(sat_per_byte_amount)

    def _get_text_from_amount(self, amount):
        amount = quantize_feerate(amount)
        return super()._get_text_from_amount(amount)



import asyncio

from PyQt6.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject

from electrum import mnemonic
from electrum import keystore
from electrum.i18n import _
from electrum.bip32 import is_bip32_derivation, xpub_type
from electrum.logging import get_logger
from electrum.util import get_asyncio_loop
from electrum.transaction import tx_from_any
from electrum.mnemonic import Mnemonic
from electrum.old_mnemonic import wordlist as old_wordlist
from electrum.bitcoin import is_address


class QEBitcoin(QObject):
    _logger = get_logger(__name__)

    generatedSeedChanged = pyqtSignal()
    seedTypeChanged = pyqtSignal()
    validationMessageChanged = pyqtSignal()

    def __init__(self, config, parent=None):
        super().__init__(parent)
        self.config = config
        self._seed_type = ''
        self._generated_seed = ''
        self._validationMessage = ''
        self._words = None

    @pyqtProperty(str, notify=generatedSeedChanged)
    def generatedSeed(self):
        return self._generated_seed

    @pyqtProperty(str, notify=seedTypeChanged)
    def seedType(self):
        return self._seed_type

    @pyqtProperty(str, notify=validationMessageChanged)
    def validationMessage(self):
        return self._validationMessage

    @validationMessage.setter
    def validationMessage(self, msg):
        if self._validationMessage != msg:
            self._validationMessage = msg
            self.validationMessageChanged.emit()

    @pyqtSlot()
    @pyqtSlot(str)
    @pyqtSlot(str, str)
    def generateSeed(self, seed_type='segwit', language='en'):
        self._logger.debug('generating seed of type ' + str(seed_type))

        async def co_gen_seed(seed_type, language):
            self._generated_seed = mnemonic.Mnemonic(language).make_seed(seed_type=seed_type)
            self._logger.debug('seed generated')
            self.generatedSeedChanged.emit()

        asyncio.run_coroutine_threadsafe(co_gen_seed(seed_type, language), get_asyncio_loop())

    @pyqtSlot(str, str, result=bool)
    def verifyMasterKey(self, key, wallet_type='standard'):
        self.validationMessage = ''
        if not keystore.is_master_key(key):
            self.validationMessage = _('Not a master key')
            return False

        k = keystore.from_master_key(key)
        if wallet_type == 'standard':
            if isinstance(k, keystore.Xpub):  # has bip32 xpub
                t1 = xpub_type(k.xpub)
                if t1 not in ['standard', 'p2wpkh', 'p2wpkh-p2sh']:  # disallow Ypub/Zpub
                    self.validationMessage = '%s: %s' % (_('Wrong key type'), t1)
                    return False
            elif isinstance(k, keystore.Old_KeyStore):
                pass
            else:
                self._logger.error(f"unexpected keystore type: {type(keystore)}")
                return False
        elif wallet_type == 'multisig':
            if not isinstance(k, keystore.Xpub):  # old mpk?
                self.validationMessage = '%s: %s' % (_('Wrong key type'), "not bip32")
                return False
            t1 = xpub_type(k.xpub)
            if t1 not in ['standard', 'p2wsh', 'p2wsh-p2sh']:  # disallow ypub/zpub
                self.validationMessage = '%s: %s' % (_('Wrong key type'), t1)
                return False
        else:
            self.validationMessage = '%s: %s' % (_('Unsupported wallet type'), wallet_type)
            self._logger.error(f'Unsupported wallet type: {wallet_type}')
            return False
        # looks okay
        return True

    @pyqtSlot(str, result=bool)
    def verifyDerivationPath(self, path):
        return is_bip32_derivation(path)

    @pyqtSlot(str, result=bool)
    def isRawTx(self, rawtx):
        try:
            tx_from_any(rawtx)
            return True
        except Exception:
            return False

    @pyqtSlot(str, result=bool)
    def isAddress(self, addr: str):
        return is_address(addr)

    @pyqtSlot(str, result=bool)
    def isAddressList(self, csv: str):
        return keystore.is_address_list(csv)

    @pyqtSlot(str, result=bool)
    def isPrivateKeyList(self, csv: str):
        return keystore.is_private_key_list(csv)

    @pyqtSlot(str, result='QVariantList')
    def mnemonicsFor(self, fragment):
        if not fragment:
            return []
        if not self._words:
            self._words = set(Mnemonic('en').wordlist).union(set(old_wordlist))
        return sorted(filter(lambda x: x.startswith(fragment), self._words))



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, List, Dict, Sequence, Set, TYPE_CHECKING
import enum
import copy

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QStandardItemModel, QStandardItem, QFont
from PyQt6.QtWidgets import QAbstractItemView, QMenu

from electrum.i18n import _
from electrum.bitcoin import is_address
from electrum.transaction import PartialTxInput, PartialTxOutput
from electrum.lnutil import MIN_FUNDING_SAT
from electrum.util import profiler
from electrum.plugin import run_hook

from .util import ColorScheme, MONOSPACE_FONT
from .my_treeview import MyTreeView, MySortModel
from .new_channel_dialog import NewChannelDialog
from ..messages import MSG_FREEZE_ADDRESS, MSG_FREEZE_COIN

if TYPE_CHECKING:
    from .main_window import ElectrumWindow


class UTXOList(MyTreeView):
    _spend_set: Set[str]  # coins selected by the user to spend from
    _utxo_dict: Dict[str, PartialTxInput]  # coin name -> coin

    class Columns(MyTreeView.BaseColumnsEnum):
        OUTPOINT = enum.auto()
        ADDRESS = enum.auto()
        LABEL = enum.auto()
        AMOUNT = enum.auto()
        PARENTS = enum.auto()

    headers = {
        Columns.OUTPOINT: _('Output point'),
        Columns.ADDRESS: _('Address'),
        Columns.PARENTS: _('Parents'),
        Columns.LABEL: _('Label'),
        Columns.AMOUNT: _('Amount'),
    }
    filter_columns = [Columns.ADDRESS, Columns.LABEL, Columns.OUTPOINT]
    stretch_column = Columns.LABEL

    ROLE_PREVOUT_STR = Qt.ItemDataRole.UserRole + 1000
    ROLE_SORT_ORDER = Qt.ItemDataRole.UserRole + 1001
    key_role = ROLE_PREVOUT_STR

    def __init__(self, main_window: 'ElectrumWindow'):
        super().__init__(
            main_window=main_window,
            stretch_column=self.stretch_column,
        )
        self._spend_set = set()
        self._utxo_dict = {}
        self.wallet = self.main_window.wallet
        self.std_model = QStandardItemModel(self)
        self.proxy = MySortModel(self, sort_role=self.ROLE_SORT_ORDER)
        self.proxy.setSourceModel(self.std_model)
        self.setModel(self.proxy)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSortingEnabled(True)

    def create_toolbar(self, config):
        toolbar, menu = self.create_toolbar_with_menu('')
        self.num_coins_label = toolbar.itemAt(0).widget()
        menu.addAction(_('Coin control'), lambda: self.add_selection_to_coincontrol())

        def cb():
            self.main_window.utxo_list.refresh_all()  # for coin frozen status
            self.main_window.update_status()  # frozen balance
        menu.addConfig(config.cv.WALLET_FREEZE_REUSED_ADDRESS_UTXOS, callback=cb)
        return toolbar

    @profiler(min_threshold=0.05)
    def update(self):
        # not calling maybe_defer_update() as it interferes with coincontrol status bar
        self.proxy.setDynamicSortFilter(False)  # temp. disable re-sorting after every change
        utxos = self.wallet.get_utxos()
        self._maybe_reset_coincontrol(utxos)
        self._utxo_dict = dict([(utxo.prevout.to_str(), utxo) for utxo in utxos])
        self.std_model.clear()
        self.update_headers(self.__class__.headers)
        for idx, utxo in enumerate(utxos):
            name = utxo.prevout.to_str()
            labels = [""] * len(self.Columns)
            amount_str = self.main_window.format_amount(
                utxo.value_sats(), whitespaces=True)
            amount_str_nots = self.main_window.format_amount(
                utxo.value_sats(), whitespaces=False, add_thousands_sep=False)
            labels[self.Columns.OUTPOINT] = str(utxo.short_id)
            labels[self.Columns.ADDRESS] = utxo.address
            labels[self.Columns.AMOUNT] = amount_str
            utxo_item = [QStandardItem(x) for x in labels]
            self.set_editability(utxo_item)
            utxo_item[self.Columns.OUTPOINT].setData(name, self.ROLE_PREVOUT_STR)
            utxo_item[self.Columns.AMOUNT].setData(amount_str_nots, self.ROLE_CLIPBOARD_DATA)
            utxo_item[self.Columns.ADDRESS].setFont(QFont(MONOSPACE_FONT))
            utxo_item[self.Columns.AMOUNT].setFont(QFont(MONOSPACE_FONT))
            utxo_item[self.Columns.PARENTS].setFont(QFont(MONOSPACE_FONT))
            utxo_item[self.Columns.OUTPOINT].setFont(QFont(MONOSPACE_FONT))
            self.std_model.insertRow(idx, utxo_item)
            self.refresh_row(name, idx)
        self.filter()
        self.proxy.setDynamicSortFilter(True)
        self.sortByColumn(self.Columns.OUTPOINT, Qt.SortOrder.DescendingOrder)
        self.update_coincontrol_bar()
        self.num_coins_label.setText(_('{} unspent transaction outputs').format(len(utxos)))

    def update_coincontrol_bar(self):
        # update coincontrol status bar
        if bool(self._spend_set):
            coins = [self._utxo_dict[x] for x in self._spend_set]
            coins = self._filter_frozen_coins(coins)
            amount = sum(x.value_sats() for x in coins)
            amount_str = self.main_window.format_amount_and_units(amount)
            num_outputs_str = _("{} outputs available ({} total)").format(len(coins), len(self._utxo_dict))
            self.main_window.set_coincontrol_msg(_("Coin control active") + f': {num_outputs_str}, {amount_str}')
        else:
            self.main_window.set_coincontrol_msg(None)

    def refresh_row(self, key, row):
        assert row is not None
        utxo = self._utxo_dict[key]
        utxo_item = [self.std_model.item(row, col) for col in self.Columns]
        txid = utxo.prevout.txid.hex()
        num_parents = self.wallet.get_num_parents(txid)
        utxo_item[self.Columns.PARENTS].setText('%6s'%num_parents if num_parents else '-')
        label = self.wallet.get_label_for_txid(txid) or ''
        utxo_item[self.Columns.LABEL].setText(label)
        sort_key = (
            self.wallet.adb.tx_height_to_sort_height(utxo.block_height),  # sort by block height
            str(utxo.short_id),                                           # order inside block (if mined), or just txid
        )
        utxo_item[self.Columns.OUTPOINT].setData(sort_key, self.ROLE_SORT_ORDER)
        SELECTED_TO_SPEND_TOOLTIP = _('Coin selected to be spent')
        if key in self._spend_set:
            tooltip = key + "\n" + SELECTED_TO_SPEND_TOOLTIP
            color = ColorScheme.GREEN.as_color(True)
        else:
            tooltip = key
            color = self._default_bg_brush
        for col in utxo_item:
            col.setBackground(color)
            col.setToolTip(tooltip)
        if self.wallet.is_frozen_address(utxo.address):
            utxo_item[self.Columns.ADDRESS].setBackground(ColorScheme.BLUE.as_color(True))
            utxo_item[self.Columns.ADDRESS].setToolTip(_('Address is frozen'))
        if self.wallet.is_frozen_coin(utxo):
            utxo_item[self.Columns.OUTPOINT].setBackground(ColorScheme.BLUE.as_color(True))
            utxo_item[self.Columns.OUTPOINT].setToolTip(f"{key}\n{_('Coin is frozen')}")

    def get_selected_outpoints(self) -> List[str]:
        if not self.model():
            return []
        items = self.selected_in_column(self.Columns.OUTPOINT)
        return [x.data(self.ROLE_PREVOUT_STR) for x in items]

    def _filter_frozen_coins(self, coins: List[PartialTxInput]) -> List[PartialTxInput]:
        coins = [utxo for utxo in coins
                 if (not self.wallet.is_frozen_address(utxo.address) and
                     not self.wallet.is_frozen_coin(utxo))]
        return coins

    def are_in_coincontrol(self, coins: List[PartialTxInput]) -> bool:
        return all([utxo.prevout.to_str() in self._spend_set for utxo in coins])

    def add_to_coincontrol(self, coins: List[PartialTxInput]):
        assert all(utxo.prevout.to_str() in self._utxo_dict for utxo in coins) # see issue 10206
        coins = self._filter_frozen_coins(coins)
        for utxo in coins:
            self._spend_set.add(utxo.prevout.to_str())
        self._refresh_coincontrol()

    def remove_from_coincontrol(self, coins: List[PartialTxInput]):
        for utxo in coins:
            self._spend_set.remove(utxo.prevout.to_str())
        self._refresh_coincontrol()

    def clear_coincontrol(self):
        self._spend_set.clear()
        self._refresh_coincontrol()

    def add_selection_to_coincontrol(self):
        if bool(self._spend_set):
            self.clear_coincontrol()
            return
        selected = self.get_selected_outpoints()
        coins = [self._utxo_dict[name] for name in selected]
        if not coins:
            self.main_window.show_error(_('You need to select coins from the list first.\nUse ctrl+left mouse button to select multiple items'))
            return
        self.add_to_coincontrol(coins)

    def _refresh_coincontrol(self):
        self.refresh_all()
        self.update_coincontrol_bar()
        self.selectionModel().clearSelection()

    def get_spend_list(self) -> Optional[Sequence[PartialTxInput]]:
        if not bool(self._spend_set):
            return None
        utxos = [self._utxo_dict[x] for x in self._spend_set]
        return copy.deepcopy(utxos)  # copy so that side-effects don't affect utxo_dict

    def _maybe_reset_coincontrol(self, current_wallet_utxos: Sequence[PartialTxInput]) -> None:
        if not self._spend_set and not self._currently_open_menu:
            return
        utxo_set = {utxo.prevout.to_str() for utxo in current_wallet_utxos}
        if self._currently_open_menu:
            # if we spent one of the qt-highlighted UTXOs, close context-menu
            if not all(prevout_str in utxo_set for prevout_str in self.get_selected_outpoints()):
                self.close_menu()
        if self._spend_set:
            # if we spent one of the green-marked UTXOs, just reset selection
            if not all([prevout_str in utxo_set for prevout_str in self._spend_set]):
                self._spend_set.clear()

    def can_swap_coins(self, coins):
        # fixme: min and max_amounts are known only after first request
        if self.wallet.lnworker is None:
            return False
        value = sum(x.value_sats() for x in coins)
        min_amount = self.wallet.lnworker.swap_manager.get_min_amount()
        max_amount = self.wallet.lnworker.swap_manager.client_max_amount_forward_swap()
        if min_amount is None or max_amount is None:
            # we need to fetch data from swap server
            return True
        if value < min_amount:
            return False
        if max_amount is None or value > max_amount:
            return False
        return True

    def swap_coins(self, coins: list[PartialTxInput]) -> None:
        assert coins, "no coins selected?"
        self.main_window.run_swap_dialog(is_reverse=False, recv_amount_sat_or_max='!', get_coins=lambda *args, **kwargs: coins)

    def can_open_channel(self, coins):
        if self.wallet.lnworker is None:
            return False
        value = sum(x.value_sats() for x in coins)
        return value >= MIN_FUNDING_SAT and value <= self.config.LIGHTNING_MAX_FUNDING_SAT

    def open_channel_with_coins(self, coins: list[PartialTxInput]) -> None:
        assert coins, "no coins selected?"
        # todo : use a single dialog in new flow
        d = NewChannelDialog(self.main_window, get_coins=lambda *args, **kwargs: coins)
        d.max_button.setChecked(True)
        d.max_button.setEnabled(False)
        d.min_button.setEnabled(False)
        d.clear_button.setEnabled(False)
        d.amount_e.setFrozen(True)
        d.spend_max()
        d.run()

    def clipboard_contains_address(self) -> bool:
        text = self.main_window.app.clipboard().text()
        return is_address(text)

    def pay_to_clipboard_address(self, coins: list[PartialTxInput]) -> None:
        assert coins, "no coins selected?"
        if not self.clipboard_contains_address():
            self.main_window.show_error(_('Clipboard doesn\'t contain a valid address'))
            return
        addr = self.main_window.app.clipboard().text()
        outputs = [PartialTxOutput.from_address_and_value(addr, '!')]

        self.main_window.send_tab.pay_onchain_dialog(outputs, get_coins=lambda *args, **kwargs: coins)

    def on_double_click(self, idx):
        outpoint = idx.sibling(idx.row(), self.Columns.OUTPOINT).data(self.ROLE_PREVOUT_STR)
        utxo = self._utxo_dict[outpoint]
        self.main_window.show_utxo(utxo)

    def create_menu(self, position):
        selected = self.get_selected_outpoints()
        coins = [self._utxo_dict[name] for name in selected]

        if not coins:
            return

        unfrozen_coins = self._filter_frozen_coins(coins)
        menu = QMenu()
        menu.setSeparatorsCollapsible(True)  # consecutive separators are merged together

        if len(coins) == 1:
            idx = self.indexAt(position)
            if not idx.isValid():
                return
            utxo = coins[0]
            txid = utxo.prevout.txid.hex()
            # "Details"
            tx = self.wallet.adb.get_transaction(txid)
            if tx:
                label = self.wallet.get_label_for_txid(txid)
                menu.addAction(_("Privacy analysis"), lambda: self.main_window.show_utxo(utxo))
            cc = self.add_copy_menu(menu, idx)
            cc.addAction(_("Long Output point"), lambda: self.place_text_on_clipboard(utxo.prevout.to_str(), title="Long Output point"))
        # fully spend
        m = menu_spend = menu.addMenu(_("Fully spend") + '…')
        m.setEnabled(bool(unfrozen_coins))
        m = menu_spend.addAction(_("send to address in clipboard"), lambda: self.pay_to_clipboard_address(unfrozen_coins))
        m.setEnabled(self.clipboard_contains_address())
        m = menu_spend.addAction(_("in new channel"), lambda: self.open_channel_with_coins(unfrozen_coins))
        m.setEnabled(self.can_open_channel(unfrozen_coins))
        m = menu_spend.addAction(_("in submarine swap"), lambda: self.swap_coins(unfrozen_coins))
        m.setEnabled(self.can_swap_coins(unfrozen_coins))
        # coin control
        if self.are_in_coincontrol(coins):
            menu.addAction(_("Remove from coin control"), lambda: self.remove_from_coincontrol(coins))
        else:
            m = menu.addAction(_("Add to coin control"), lambda: self.add_to_coincontrol(coins))
            m.setEnabled(bool(unfrozen_coins))
        # Freeze menu
        if len(coins) == 1:
            utxo = coins[0]
            addr = utxo.address
            menu_freeze = menu.addMenu(_("Freeze"))
            menu_freeze.setToolTipsVisible(True)
            if not self.wallet.is_frozen_coin(utxo):
                act = menu_freeze.addAction(_("Freeze Coin"), lambda: self.main_window.set_frozen_state_of_coins([utxo], True))
            else:
                act = menu_freeze.addAction(_("Unfreeze Coin"), lambda: self.main_window.set_frozen_state_of_coins([utxo], False))
            act.setToolTip(MSG_FREEZE_COIN)
            if not self.wallet.is_frozen_address(addr):
                act = menu_freeze.addAction(_("Freeze Address"), lambda: self.main_window.set_frozen_state_of_addresses([addr], True))
            else:
                act = menu_freeze.addAction(_("Unfreeze Address"), lambda: self.main_window.set_frozen_state_of_addresses([addr], False))
            act.setToolTip(MSG_FREEZE_ADDRESS)
        elif len(coins) > 1:  # multiple items selected
            menu.addSeparator()
            addrs = [utxo.address for utxo in coins]
            is_coin_frozen = [self.wallet.is_frozen_coin(utxo) for utxo in coins]
            is_addr_frozen = [self.wallet.is_frozen_address(utxo.address) for utxo in coins]
            menu_freeze = menu.addMenu(_("Freeze"))
            menu_freeze.setToolTipsVisible(True)
            if not all(is_coin_frozen):
                act = menu_freeze.addAction(_("Freeze Coins"), lambda: self.main_window.set_frozen_state_of_coins(coins, True))
                act.setToolTip(MSG_FREEZE_COIN)
            if any(is_coin_frozen):
                act = menu_freeze.addAction(_("Unfreeze Coins"), lambda: self.main_window.set_frozen_state_of_coins(coins, False))
                act.setToolTip(MSG_FREEZE_COIN)
            if not all(is_addr_frozen):
                act = menu_freeze.addAction(_("Freeze Addresses"), lambda: self.main_window.set_frozen_state_of_addresses(addrs, True))
                act.setToolTip(MSG_FREEZE_ADDRESS)
            if any(is_addr_frozen):
                act = menu_freeze.addAction(_("Unfreeze Addresses"), lambda: self.main_window.set_frozen_state_of_addresses(addrs, False))
                act.setToolTip(MSG_FREEZE_ADDRESS)

        run_hook('qt_utxo_menu', menu, coins, self.wallet)
        self.open_menu(menu, position)

    def get_filter_data_from_coordinate(self, row, col):
        if col == self.Columns.OUTPOINT:
            return self.get_role_data_from_coordinate(row, col, role=self.ROLE_PREVOUT_STR)
        return super().get_filter_data_from_coordinate(row, col)



# Electrum - lightweight Bitcoin client
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
import json
import locale
import traceback
import sys
import queue
from typing import TYPE_CHECKING, NamedTuple, Optional, TypedDict
from types import TracebackType

from .version import ELECTRUM_VERSION
from . import constants
from .i18n import _
from .util import make_aiohttp_session, error_text_str_to_safe_str
from .logging import describe_os_version, Logger, get_git_version
from .crypto import sha256

if TYPE_CHECKING:
    from .network import ProxySettings


_tainted_by_console = False
def taint_reports_by_console_usage():
    global _tainted_by_console
    _tainted_by_console = True


class CrashReportResponse(NamedTuple):
    status: Optional[str]
    text: str
    url: Optional[str]


class BaseCrashReporter(Logger):
    report_server = "https://crashhub.electrum.org"
    issue_template = """<h2>Traceback</h2>
<pre>
{traceback}
</pre>

<h2>Additional information</h2>
<ul>
  <li>Electrum version: {app_version}</li>
  <li>Python version: {python_version}</li>
  <li>Operating system: {os}</li>
  <li>Wallet type: {wallet_type}</li>
  <li>Locale: {locale}</li>
</ul>
    """
    CRASH_MESSAGE = _('Something went wrong while executing Electrum.')
    CRASH_TITLE = _('Sorry!')
    REQUEST_HELP_MESSAGE = _('To help us diagnose and fix the problem, you can send us a bug report that contains '
                             'useful debug information:')
    DESCRIBE_ERROR_MESSAGE = _("Please briefly describe what led to the error (optional):")
    ASK_CONFIRM_SEND = _("Do you want to send this report?")
    USER_COMMENT_PLACEHOLDER = _("Do not enter sensitive/private information here. "
                                 "The report will be visible on the public issue tracker.")

    exc_args: tuple[type[BaseException], BaseException, TracebackType | None]

    def __init__(
        self,
        exctype: type[BaseException],
        excvalue: BaseException,
        tb: TracebackType | None,
    ):
        Logger.__init__(self)
        self.exc_args = (exctype, excvalue, tb)

    def send_report(self, asyncio_loop, proxy: 'ProxySettings', *, timeout=None) -> CrashReportResponse:
        # FIXME the caller needs to catch generic "Exception", as this method does not have a well-defined API...
        bitcoinlike_genesis = constants.net.GENESIS[-4:] in [
            "e26f",  # mainnet
            "4943",  # testnet 3
            "f043",  # testnet 4
            "1ef6",  # signet
        ]
        is_forked_codebase = not bitcoinlike_genesis or "/spesmilo/" not in constants.GIT_REPO_URL
        if is_forked_codebase and ".electrum.org" in BaseCrashReporter.report_server:
            # Some kind of altcoin wants to send us crash reports... ?
            # - You, the reader, yes *you*: if you forked the codebase e.g. to support an altcoin,
            #   please consider that upstream Electrum does NOT want to receive automated crash reports
            #   from your users.  You should set up your own crash report server,
            #   and change `BaseCrashReporter.report_server` accordingly, or just disable reporting.
            raise Exception(_("Missing report URL."))
        report = self.get_traceback_info(*self.exc_args)
        report.update(self.get_additional_info())
        report = json.dumps(report)
        coro = self.do_post(proxy, BaseCrashReporter.report_server + "/crash.json", data=report)
        response = asyncio.run_coroutine_threadsafe(coro, asyncio_loop).result(timeout)
        self.logger.info(
            f"Crash report sent. Got response [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(response)}")
        response = json.loads(response)
        assert isinstance(response, dict), type(response)
        # sanitize URL
        if location := response.get("location"):
            assert isinstance(location, str)
            base_issues_url = constants.GIT_REPO_ISSUES_URL
            if not base_issues_url.endswith("/"):
                base_issues_url = base_issues_url + "/"
            if not location.startswith(base_issues_url):
                location = None
        ret = CrashReportResponse(
            status=response.get("status"),
            url=location,
            text=_("Thanks for reporting this issue!"),
        )
        return ret

    async def do_post(self, proxy: 'ProxySettings', url, data) -> str:
        async with make_aiohttp_session(proxy) as session:
            async with session.post(url, data=data, raise_for_status=True) as resp:
                return await resp.text()

    @classmethod
    def get_traceback_info(
        cls,
        exctype: type[BaseException],
        excvalue: BaseException,
        tb: TracebackType | None,
    ) -> TypedDict('TBInfo', {'exc_string': str, 'stack': str, 'id': dict[str, str]}):
        exc_string = str(excvalue)
        stack = traceback.extract_tb(tb)
        readable_trace = cls._get_traceback_str_to_send(exctype, excvalue, tb)
        _id = {
            "file": stack[-1].filename if len(stack) else '<no stack>',
            "name": stack[-1].name if len(stack) else '<no stack>',
            "type": exctype.__name__
        }  # note: this is the "id" the crash reporter server uses to group together reports.
        return {
            "exc_string": exc_string,
            "stack": readable_trace,
            "id": _id,
        }

    @classmethod
    def get_traceback_groupid_hash(
        cls,
        exctype: type[BaseException],
        excvalue: BaseException,
        tb: TracebackType | None,
    ) -> bytes:
        tb_info = cls.get_traceback_info(exctype, excvalue, tb)
        _id = tb_info["id"]
        return sha256(str(_id))

    def get_additional_info(self):
        app_version = (get_git_version() or ELECTRUM_VERSION)
        if _tainted_by_console:
            app_version += "-consoletaint"
        args = {
            "app_version": app_version,
            "python_version": sys.version,
            "os": describe_os_version(),
            "wallet_type": "unknown",
            "locale": locale.getlocale()[0] or "?",
            "description": self.get_user_description()
        }
        try:
            args["wallet_type"] = self.get_wallet_type()
        except Exception:
            # Maybe the wallet isn't loaded yet
            pass
        return args

    @classmethod
    def _get_traceback_str_to_send(
        cls,
        exctype: type[BaseException],
        excvalue: BaseException,
        tb: TracebackType | None,
    ) -> str:
        # make sure that traceback sent to crash reporter contains
        # e.__context__ and e.__cause__, i.e. if there was a chain of
        # exceptions, we want the full traceback for the whole chain.
        return "".join(traceback.format_exception(exctype, excvalue, tb))

    def _get_traceback_str_to_display(self) -> str:
        # overridden in Qt subclass
        return self._get_traceback_str_to_send(*self.exc_args)

    def get_report_string(self):
        info = self.get_additional_info()
        info["traceback"] = self._get_traceback_str_to_display()
        return self.issue_template.format(**info)

    def get_user_description(self):
        raise NotImplementedError

    def get_wallet_type(self) -> str:
        raise NotImplementedError


class EarlyExceptionsQueue:
    """Helper singleton for explicitly sending exceptions to crash reporter.

    Typically the GUIs set up an "exception hook" that catches all otherwise
    uncaught exceptions (which unroll the stack of a thread completely).
    This class provides methods to report *any* exception, and queueing logic
    that delays processing until the exception hook is set up.
    """

    _is_exc_hook_ready = False
    _exc_queue = queue.Queue()

    @classmethod
    def set_hook_as_ready(cls):
        """Flush the queue and disable it for future exceptions."""
        if cls._is_exc_hook_ready:
            return
        cls._is_exc_hook_ready = True
        while cls._exc_queue.qsize() > 0:
            e = cls._exc_queue.get()
            cls._send_exception_to_crash_reporter(e)

    @classmethod
    def send_exception_to_crash_reporter(cls, e: BaseException):
        if cls._is_exc_hook_ready:
            cls._send_exception_to_crash_reporter(e)
        else:
            cls._exc_queue.put(e)

    @staticmethod
    def _send_exception_to_crash_reporter(e: BaseException):
        assert EarlyExceptionsQueue._is_exc_hook_ready
        sys.excepthook(type(e), e, e.__traceback__)


send_exception_to_crash_reporter = EarlyExceptionsQueue.send_exception_to_crash_reporter


def trigger_crash():
    # note: do not change the type of the exception, the message,
    # or the name of this method. All reports generated through this
    # method will be grouped together by the crash reporter, and thus
    # don't spam the issue tracker.

    class TestingException(Exception):
        pass

    def crash_test():
        raise TestingException("triggered crash for testing purposes")

    import threading
    t = threading.Thread(target=crash_test)
    t.start()
